@@ -1,9 +1,7 @@
-/*
- * SPDX-License-Identifier: GPL-2.0-only
- *
- * Copyright (C) 2011 Julien Desfossez <julien.desfossez@polymtl.ca>
- * Copyright (C) 2011-2013 Mathieu Desnoyers <mathieu.desnoyers@efficios.com>
- */
+// SPDX-FileCopyrightText: 2011 EfficiOS Inc.
+// SPDX-FileCopyrightText: 2011-2013 Mathieu Desnoyers <mathieu.desnoyers@efficios.com>
+//
+// SPDX-License-Identifier: GPL-2.0-only
 
 #ifndef _LTTNG_UST_CTL_H
 #define _LTTNG_UST_CTL_H
@@ -22,10 +20,6 @@
 extern "C" {
 #endif
 
-#ifndef LTTNG_UST_UUID_LEN
-#define LTTNG_UST_UUID_LEN	16
-#endif
-
 /* Default unix socket path */
 #define LTTNG_UST_SOCK_FILENAME					\
 	"lttng-ust-sock-"					\
@@ -48,8 +42,11 @@ struct lttng_ust_ctl_consumer_channel_attr {
 	unsigned int read_timer_interval;	/* usec */
 	enum lttng_ust_abi_output output;	/* splice, mmap */
 	uint32_t chan_id;			/* channel ID */
-	unsigned char uuid[LTTNG_UST_UUID_LEN]; /* Trace session unique ID */
+	unsigned char uuid[LTTNG_UST_ABI_UUID_LEN]; /* Trace session unique ID */
 	int64_t blocking_timeout;			/* Blocking timeout (usec) */
+	uint32_t owner_id;
+	/* 1: preallocate/zero, 0: don't preallocate (sparse/lazy allocation) */
+	int preallocate_backing;
 } __attribute__((packed));
 
 /*
@@ -143,6 +140,9 @@ int lttng_ust_ctl_tracepoint_field_list_get(int sock, int tp_field_list_handle,
 		struct lttng_ust_abi_field_iter *iter);
 
 int lttng_ust_ctl_tracer_version(int sock, struct lttng_ust_abi_tracer_version *v);
+
+int lttng_ust_ctl_unknown_command(int sock);
+
 int lttng_ust_ctl_wait_quiescent(int sock);
 
 int lttng_ust_ctl_sock_flush_buffer(int sock, struct lttng_ust_abi_object_data *object);
@@ -179,6 +179,7 @@ int lttng_ust_ctl_duplicate_ust_object_data(struct lttng_ust_abi_object_data **d
 struct lttng_ust_ctl_consumer_channel;
 struct lttng_ust_ctl_consumer_stream;
 struct lttng_ust_ctl_consumer_channel_attr;
+struct lttng_ust_ctl_consumer_packet;
 
 int lttng_ust_ctl_get_nr_stream_per_channel(void);
 
@@ -229,6 +230,149 @@ int lttng_ust_ctl_get_mmap_len(struct lttng_ust_ctl_consumer_stream *stream,
 		unsigned long *len);
 int lttng_ust_ctl_get_max_subbuf_size(struct lttng_ust_ctl_consumer_stream *stream,
 		unsigned long *len);
+int lttng_ust_ctl_stream_get_backend_area(struct lttng_ust_ctl_consumer_stream *stream,
+		void **addr, unsigned long *len);
+
+void lttng_ust_ctl_set_channel_owner_id(struct lttng_ust_abi_object_data *obj,
+					uint32_t id);
+
+/* Opaque type. */
+struct lttng_ust_ctl_subbuf_iter;
+
+enum lttng_ust_ctl_subbuf_iter_type {
+	/* Iterate on all delivered subbuffers: those non-consumed and consumed. */
+	LTTNG_UST_CTL_SUBBUF_ITER_DELIVERED = 0,
+	/* Iterate on subbuffers which were delivered and consumed. */
+	LTTNG_UST_CTL_SUBBUF_ITER_DELIVERED_CONSUMED = 1,
+	/* Iterate only on non-consumed subbuffers, whether they were delivered or not. */
+	LTTNG_UST_CTL_SUBBUF_ITER_UNCONSUMED = 2,
+};
+
+/*
+ * Create a subbuffer iterator object.
+ *
+ * The object will iterate on sub-buffers from `stream`.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_create(struct lttng_ust_ctl_consumer_stream *stream,
+				enum lttng_ust_ctl_subbuf_iter_type iter_type,
+				struct lttng_ust_ctl_subbuf_iter **iter);
+
+/*
+ * Iterate to the next sub-buffer.
+ *
+ * Return 0 if no next sub-buffer is available.
+ *
+ * Return 1 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_next(struct lttng_ust_ctl_subbuf_iter *iter);
+
+
+/*
+ * Get the hot commit counter of the current sub-buffer.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_cc_hot(struct lttng_ust_ctl_subbuf_iter *iter,
+				unsigned long *cc_hot);
+
+/*
+ * Get the cold commit counter of the current sub-buffer.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_cc_cold(struct lttng_ust_ctl_subbuf_iter *iter,
+				unsigned long *cc_cold);
+
+/*
+ * Get the index in the buffer of the current sub-buffer.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_index(struct lttng_ust_ctl_subbuf_iter *iter,
+				size_t *idx);
+
+/*
+ * Get the owner of the current sub-buffer.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_owner(struct lttng_ust_ctl_subbuf_iter *iter,
+				uint32_t *owner);
+
+/*
+ * Get the timestamp begin of the current sub-buffer.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_timestamp_begin(struct lttng_ust_ctl_subbuf_iter *iter,
+				uint64_t *timestamp_begin);
+
+/*
+ * Get the timestamp end of the current sub-buffer.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_timestamp_end(struct lttng_ust_ctl_subbuf_iter *iter,
+				uint64_t *timestamp_end);
+
+/*
+ * Get the allocated state of the current sub-buffer.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_allocated(struct lttng_ust_ctl_subbuf_iter *iter,
+				bool *allocated);
+
+/*
+ * Get the free running position of the current sub-buffer.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_pos(struct lttng_ust_ctl_subbuf_iter *iter,
+				unsigned long *pos);
+
+/*
+ * Destroy the subbuf iterator object.
+ *
+ * Return 0 on success.
+ *
+ * Return a negative value on error.
+ */
+int lttng_ust_ctl_subbuf_iter_destroy(struct lttng_ust_ctl_subbuf_iter *iter);
+
+/*
+ * Try to fixup any stalled sub-buffer in the range [consumed_pos, produced_pos].
+ *
+ * `ids[length]` is a list of known owners that are dead. Only sub-buffer owned
+ * by dead owners (or without any owner) can be fixed up by this function.
+ */
+int lttng_ust_ctl_fixup_stalled_stream(struct lttng_ust_ctl_consumer_stream *stream,
+				unsigned long consumed_pos,
+				unsigned long produced_pos,
+				uint32_t *ids,
+				uint32_t length);
 
 /*
  * For mmap mode, operate on the current packet (between get/put or
@@ -257,8 +401,28 @@ int lttng_ust_ctl_get_subbuf(struct lttng_ust_ctl_consumer_stream *stream,
 int lttng_ust_ctl_put_subbuf(struct lttng_ust_ctl_consumer_stream *stream);
 
 int lttng_ust_ctl_flush_buffer(struct lttng_ust_ctl_consumer_stream *stream,
-		int producer_active);
+			int producer_active);
+int lttng_ust_ctl_flush_buffer_with_old_position(struct lttng_ust_ctl_consumer_stream *stream,
+			int producer_active, unsigned long *old_pos);
+/*
+ * Perform an active flush, populating a packet if there was not packet
+ * delivery performed during the flush so readers can be able to infer the lack
+ * events between the last produced packet and the time the produced position
+ * sampled (indicated by an empty packet).
+ *
+ * `packet` and `packet_populated` must be non-NULL.
+ * `packet` should be a pointer to memory allocated by `lttng_ust_ctl_packet_create`.
+ */
+int lttng_ust_ctl_flush_events_or_populate_packet(struct lttng_ust_ctl_consumer_stream *stream,
+		struct lttng_ust_ctl_consumer_packet *packet,
+		bool *packet_populated, bool *flush_done);
 int lttng_ust_ctl_clear_buffer(struct lttng_ust_ctl_consumer_stream *stream);
+/*
+ * Synchronize time for this stream so no packet header nor event can be
+ * produced with a timestamp prior to the current time after this
+ * function returns.
+ */
+int lttng_ust_ctl_timestamp_sync(struct lttng_ust_ctl_consumer_stream *stream);
 
 /* index */
 
@@ -297,6 +461,59 @@ int lttng_ust_ctl_get_instance_id(struct lttng_ust_ctl_consumer_stream *stream,
 int lttng_ust_ctl_get_current_timestamp(struct lttng_ust_ctl_consumer_stream *stream,
 		uint64_t *ts);
 
+/*
+ * Return < 0 if the last activity timestamp for @stream is before @ts,
+ * return > 0 if it after, 0 if equal. Note that on 32-bit architectures
+ * the accuracy of the comparison has a coarseness in the order of
+ * 250ms.
+ *
+ * The @ts parameter is expressed in clock cycles, as defined by the
+ * clock frequency.
+ */
+int lttng_ust_ctl_last_activity_timestamp_compare(struct lttng_ust_ctl_consumer_stream *stream,
+		uint64_t ts);
+
+/*
+ * Add a nanoseconds @delta_ns (positive or negative) to a stream @timestamp.
+ * Return 0 on success, negative error value on error. @timestamp is an
+ * inout parameter.
+ *
+ * This helper takes into account the stream's clocksource frequency.
+ */
+int lttng_ust_ctl_timestamp_add(struct lttng_ust_ctl_consumer_stream *stream,
+		uint64_t *ts, int64_t delta_ns);
+
+/*
+ * Try to exchange the subbuffer corresponding to @pos with the reader subbuffer.
+ *
+ * Return -ENOENT if the subbuffer is being used for writing or there is
+ * no subbuffer matching the @pos producer position in the buffer, else
+ * return 0 on success. Return -EIO on shared memory access error.
+ */
+int lttng_ust_ctl_try_exchange_subbuf(struct lttng_ust_ctl_consumer_stream *stream, unsigned long pos);
+
+/*
+ * Reclaim memory (MADV_REMOVE) backing reader subbuffer.
+ *
+ * Return 0 on success, -ENOMEM if memory is already reclaimed. It can
+ * also return any negative errno from madvise(2).
+ * Return -EIO on shared memory access error.
+ */
+int lttng_ust_ctl_reclaim_reader_subbuf(struct lttng_ust_ctl_consumer_stream *stream);
+
+/*
+ * Packets
+ */
+
+/*
+ * Allocates and creates a new packet, owned by the caller.
+ * The packet must be destroyed using `lttng_ust_ctl_packet_destroy`.
+ */
+int lttng_ust_ctl_packet_create(struct lttng_ust_ctl_consumer_packet **packet);
+void lttng_ust_ctl_packet_destroy(struct lttng_ust_ctl_consumer_packet *packet);
+int lttng_ust_ctl_packet_get_buffer(struct lttng_ust_ctl_consumer_packet *packet, void **buffer,
+		uint64_t *packet_length, uint64_t *packet_length_padded);
+
 /* returns whether UST has perf counters support. */
 int lttng_ust_ctl_has_perf_counters(void);
 
@@ -314,6 +531,7 @@ enum lttng_ust_ctl_notify_cmd {
 	LTTNG_UST_CTL_NOTIFY_CMD_EVENT = 0,
 	LTTNG_UST_CTL_NOTIFY_CMD_CHANNEL = 1,
 	LTTNG_UST_CTL_NOTIFY_CMD_ENUM = 2,
+	LTTNG_UST_CTL_NOTIFY_CMD_KEY = 3,
 };
 
 enum lttng_ust_ctl_channel_header {
@@ -538,15 +756,54 @@ int lttng_ust_ctl_recv_register_event(int sock,
 					 */
 	size_t *nr_fields,
 	struct lttng_ust_ctl_field **fields,
-	char **model_emf_uri);
+	char **model_emf_uri,
+	uint64_t *user_token);
 
 /*
  * Returns 0 on success, negative error value on error.
  */
 int lttng_ust_ctl_reply_register_event(int sock,
-	uint32_t id,			/* event id (input) */
+	uint32_t id,			/* id (input) */
 	int ret_code);			/* return code. 0 ok, negative error */
 
+#ifdef CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER
+/*
+ * Returns 0 on success, negative UST or system error value on error.
+ */
+int lttng_ust_ctl_recv_register_key(int sock,
+	int *session_objd,		/* session descriptor (output) */
+	int *map_objd,			/* map descriptor (output) */
+	uint32_t *dimension,		/*
+					 * Against which dimension is
+					 * this key expressed. (output)
+					 */
+	uint64_t **dimension_indexes,	/*
+					 * Indexes (output,
+					 * dynamically
+					 * allocated, must be
+					 * free(3)'d by the
+					 * caller if function
+					 * returns success.)
+					 * Contains @dimension
+					 * elements.
+					 */
+	char **key_string,		/*
+					 * key string (output,
+					 * dynamically allocated, must
+					 * be free(3)'d by the caller if
+					 * function returns success.)
+					 */
+	uint64_t *user_token);
+
+/*
+ * Returns 0 on success, negative error value on error.
+ */
+int lttng_ust_ctl_reply_register_key(int sock,
+	uint64_t index,			/* Index within dimension (input) */
+	int ret_code);			/* return code. 0 ok, negative error */
+
+#endif	/* CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER */
+
 /*
  * Returns 0 on success, negative UST or system error value on error.
  */
@@ -594,10 +851,15 @@ enum lttng_ust_ctl_counter_arithmetic {
 	LTTNG_UST_CTL_COUNTER_ARITHMETIC_SATURATION	= 1,
 };
 
+enum lttng_ust_ctl_key_type {
+	LTTNG_UST_CTL_KEY_TYPE_TOKENS = 0,
+	LTTNG_UST_CTL_KEY_TYPE_INTEGER = 1,
+};
+
 /* Used as alloc flags. */
 enum lttng_ust_ctl_counter_alloc {
 	LTTNG_UST_CTL_COUNTER_ALLOC_PER_CPU = (1 << 0),
-	LTTNG_UST_CTL_COUNTER_ALLOC_GLOBAL = (1 << 1),
+	LTTNG_UST_CTL_COUNTER_ALLOC_PER_CHANNEL = (1 << 1),
 };
 
 struct lttng_ust_ctl_daemon_counter;
@@ -608,6 +870,7 @@ struct lttng_ust_ctl_counter_dimension {
 	uint64_t size;
 	uint64_t underflow_index;
 	uint64_t overflow_index;
+	enum lttng_ust_ctl_key_type key_type;
 	uint8_t has_underflow;
 	uint8_t has_overflow;
 };
@@ -616,7 +879,7 @@ struct lttng_ust_ctl_daemon_counter *
 	lttng_ust_ctl_create_counter(size_t nr_dimensions,
 		const struct lttng_ust_ctl_counter_dimension *dimensions,
 		int64_t global_sum_step,
-		int global_counter_fd,
+		int per_channel_counter_fd,
 		int nr_counter_cpu_fds,
 		const int *counter_cpu_fds,
 		enum lttng_ust_ctl_counter_bitness bitness,
@@ -627,8 +890,10 @@ struct lttng_ust_ctl_daemon_counter *
 int lttng_ust_ctl_create_counter_data(struct lttng_ust_ctl_daemon_counter *counter,
 		struct lttng_ust_abi_object_data **counter_data);
 
-int lttng_ust_ctl_create_counter_global_data(struct lttng_ust_ctl_daemon_counter *counter,
-		struct lttng_ust_abi_object_data **counter_global_data);
+#ifdef CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER
+int lttng_ust_ctl_create_counter_channel_data(struct lttng_ust_ctl_daemon_counter *counter,
+		struct lttng_ust_abi_object_data **counter_channel_data);
+#endif	/* CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER */
 int lttng_ust_ctl_create_counter_cpu_data(struct lttng_ust_ctl_daemon_counter *counter, int cpu,
 		struct lttng_ust_abi_object_data **counter_cpu_data);
 
@@ -640,17 +905,21 @@ void lttng_ust_ctl_destroy_counter(struct lttng_ust_ctl_daemon_counter *counter)
 
 int lttng_ust_ctl_send_counter_data_to_ust(int sock, int parent_handle,
 		struct lttng_ust_abi_object_data *counter_data);
-int lttng_ust_ctl_send_counter_global_data_to_ust(int sock,
+#ifdef CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER
+int lttng_ust_ctl_send_counter_channel_data_to_ust(int sock,
 		struct lttng_ust_abi_object_data *counter_data,
-		struct lttng_ust_abi_object_data *counter_global_data);
+		struct lttng_ust_abi_object_data *counter_channel_data);
+#endif	/* CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER */
 int lttng_ust_ctl_send_counter_cpu_data_to_ust(int sock,
 		struct lttng_ust_abi_object_data *counter_data,
 		struct lttng_ust_abi_object_data *counter_cpu_data);
 
+#ifdef CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER
 int lttng_ust_ctl_counter_read(struct lttng_ust_ctl_daemon_counter *counter,
 		const size_t *dimension_indexes,
 		int cpu, int64_t *value,
 		bool *overflow, bool *underflow);
+#endif	/* CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER */
 int lttng_ust_ctl_counter_aggregate(struct lttng_ust_ctl_daemon_counter *counter,
 		const size_t *dimension_indexes,
 		int64_t *value,
@@ -658,8 +927,18 @@ int lttng_ust_ctl_counter_aggregate(struct lttng_ust_ctl_daemon_counter *counter
 int lttng_ust_ctl_counter_clear(struct lttng_ust_ctl_daemon_counter *counter,
 		const size_t *dimension_indexes);
 
+#ifdef CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER
+int lttng_ust_ctl_counter_create_event(int sock,
+		struct lttng_ust_abi_counter_event *counter_event,
+		size_t counter_event_len,
+		struct lttng_ust_abi_object_data *counter_data,
+		struct lttng_ust_abi_object_data **counter_event_data);
+#endif	/* CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER */
+
 void lttng_ust_ctl_sigbus_handle(void *addr);
 
+int lttng_ust_ctl_get_version(uint32_t *major, uint32_t *minor, uint32_t *patchlevel);
+
 #ifdef __cplusplus
 }
 #endif
