@@ -18,13 +18,6 @@
 #include "common/macros.h"
 #include "common/ust-context-provider.h"
 
-/*
- * The context procname length is part of the LTTng-UST ABI.
- * TODO: At the next breaking protocol bump, all users of this macro
- * should instead use LTTNG_UST_ABI_PROCNAME_LEN.
- */
-#define LTTNG_UST_CONTEXT_PROCNAME_LEN	17
-
 struct lttng_ust_abi_obj;
 struct lttng_event_notifier_group;
 
@@ -396,7 +389,7 @@ struct lttng_ust_session_private {
 	unsigned int tstate:1;			/* Transient enable state */
 	unsigned int statedump_pending:1;
 	struct lttng_ust_ctx *ctx;		/* contexts for filters. */
-	unsigned char uuid[LTTNG_UST_UUID_LEN];	/* Trace session unique ID */
+	unsigned char uuid[LTTNG_UST_ABI_UUID_LEN]; /* Trace session unique ID */
 	bool uuid_set;				/* Is uuid set ? */
 };
 
@@ -421,7 +414,10 @@ struct lttng_ust_channel_buffer_ops_private {
 			unsigned char *uuid,
 			uint32_t chan_id,
 			const int *stream_fds, int nr_stream_fds,
-			int64_t blocking_timeout);
+			int64_t blocking_timeout,
+			uint32_t owner_id,
+			bool preallocate_backing);
+
 	void (*channel_destroy)(struct lttng_ust_channel_buffer *chan);
 	/*
 	 * packet_avail_size returns the available size in the current
@@ -452,7 +448,7 @@ struct lttng_ust_channel_buffer_private {
 	enum lttng_ust_abi_chan_type type;
 	struct lttng_ust_ctx *ctx;
 	struct lttng_ust_ring_buffer_channel *rb_chan;	/* Ring buffer channel */
-	unsigned char uuid[LTTNG_UST_UUID_LEN];	/* Trace session unique ID */
+	unsigned char uuid[LTTNG_UST_ABI_UUID_LEN];	/* Trace session unique ID */
 };
 
 struct lttng_ust_channel_counter_ops_private {
@@ -495,35 +491,6 @@ struct lttng_ust_channel_counter_private {
 	enum lttng_key_type dimension_key_types[LTTNG_COUNTER_DIMENSION_MAX];
 };
 
-/*
- * IMPORTANT: this structure is part of the ABI between the consumer
- * daemon and the UST library within traced applications. Changing it
- * breaks the UST communication protocol.
- *
- * TODO: remove unused fields on next UST communication protocol
- * breaking update.
- */
-struct lttng_ust_abi_channel_config {
-	void *unused1;
-	int unused2;
-	void *unused3;
-	void *unused4;
-	int unused5;
-	unsigned int _deprecated1;
-	unsigned int _deprecated2;
-	struct cds_list_head unused6;
-	void *unused7;
-	int unused8;
-	void *unused9;
-	unsigned int _deprecated3:1;
-
-	/* Channel ID */
-	unsigned int id;
-	enum lttng_ust_abi_chan_type unused10;
-	unsigned char uuid[LTTNG_UST_UUID_LEN]; /* Trace session unique ID */
-	int unused11:1;
-};
-
 /* Global (filter), event and channel contexts. */
 struct lttng_ust_ctx {
 	struct lttng_ust_ctx_field *fields;
