@@ -20,7 +20,7 @@
 
 struct metadata_packet_header {
 	uint32_t magic;			/* 0x75D11D57 */
-	uint8_t  uuid[LTTNG_UST_UUID_LEN]; /* Unique Universal Identifier */
+	uint8_t  uuid[LTTNG_UST_ABI_UUID_LEN]; /* Unique Universal Identifier */
 	uint32_t checksum;		/* 0 if unused */
 	uint32_t content_size;		/* in bits */
 	uint32_t packet_size;		/* in bits */
@@ -93,7 +93,8 @@ static size_t client_packet_header_size(void)
 }
 
 static void client_buffer_begin(struct lttng_ust_ring_buffer *buf,
-		uint64_t tsc __attribute__((unused)),
+		uint64_t timestamp __attribute__((unused)),
+		uint64_t discarded_events __attribute__((unused)),
 		unsigned int subbuf_idx,
 		struct lttng_ust_shm_handle *handle)
 {
@@ -125,7 +126,7 @@ static void client_buffer_begin(struct lttng_ust_ring_buffer *buf,
  * subbuffer. data_size is between 1 and subbuf_size.
  */
 static void client_buffer_end(struct lttng_ust_ring_buffer *buf,
-		uint64_t tsc  __attribute__((unused)),
+		uint64_t timestamp  __attribute__((unused)),
 		unsigned int subbuf_idx, unsigned long data_size,
 		struct lttng_ust_shm_handle *handle,
 		const struct lttng_ust_ring_buffer_ctx *ctx)
@@ -137,12 +138,15 @@ static void client_buffer_end(struct lttng_ust_ring_buffer *buf,
 				subbuf_idx * chan->backend.subbuf_size,
 				handle);
 	unsigned long records_lost = 0;
+	ssize_t page_size = LTTNG_UST_PAGE_SIZE;
 
 	assert(header);
 	if (!header)
 		return;
+	if (page_size < 0)
+		return;
 	header->content_size = data_size * CHAR_BIT;		/* in bits */
-	header->packet_size = LTTNG_UST_PAGE_ALIGN(data_size) * CHAR_BIT; /* in bits */
+	header->packet_size = LTTNG_UST_ALIGN(data_size, page_size) * CHAR_BIT; /* in bits */
 	/*
 	 * We do not care about the records lost count, because the metadata
 	 * channel waits and retry.
@@ -193,9 +197,9 @@ static const struct lttng_ust_ring_buffer_config client_config = {
 	.cb.buffer_create = client_buffer_create,
 	.cb.buffer_finalize = client_buffer_finalize,
 
-	.tsc_bits = 0,
-	.alloc = RING_BUFFER_ALLOC_GLOBAL,
-	.sync = RING_BUFFER_SYNC_GLOBAL,
+	.timestamp_bits = 0,
+	.alloc = RING_BUFFER_ALLOC_PER_CHANNEL,
+	.sync = RING_BUFFER_SYNC_PER_CHANNEL,
 	.mode = RING_BUFFER_MODE_TEMPLATE,
 	.backend = RING_BUFFER_PAGE,
 	.output = RING_BUFFER_MMAP,
@@ -216,7 +220,9 @@ struct lttng_ust_channel_buffer *_channel_create(const char *name,
 				unsigned char *uuid,
 				uint32_t chan_id,
 				const int *stream_fds, int nr_stream_fds,
-				int64_t blocking_timeout)
+				int64_t blocking_timeout,
+				uint32_t owner_id,
+				bool preallocate_backing)
 {
 	struct lttng_ust_abi_channel_config chan_priv_init;
 	struct lttng_ust_shm_handle *handle;
@@ -225,11 +231,11 @@ struct lttng_ust_channel_buffer *_channel_create(const char *name,
 	lttng_chan_buf = lttng_ust_alloc_channel_buffer();
 	if (!lttng_chan_buf)
 		return NULL;
-	memcpy(lttng_chan_buf->priv->uuid, uuid, LTTNG_UST_UUID_LEN);
+	memcpy(lttng_chan_buf->priv->uuid, uuid, LTTNG_UST_ABI_UUID_LEN);
 	lttng_chan_buf->priv->id = chan_id;
 
 	memset(&chan_priv_init, 0, sizeof(chan_priv_init));
-	memcpy(chan_priv_init.uuid, uuid, LTTNG_UST_UUID_LEN);
+	memcpy(chan_priv_init.uuid, uuid, LTTNG_UST_ABI_UUID_LEN);
 	chan_priv_init.id = chan_id;
 
 	handle = channel_create(&client_config, name,
@@ -238,7 +244,8 @@ struct lttng_ust_channel_buffer *_channel_create(const char *name,
 			&chan_priv_init,
 			lttng_chan_buf, buf_addr, subbuf_size, num_subbuf,
 			switch_timer_interval, read_timer_interval,
-			stream_fds, nr_stream_fds, blocking_timeout);
+			stream_fds, nr_stream_fds, blocking_timeout,
+			owner_id, preallocate_backing);
 	if (!handle)
 		goto error;
 	lttng_chan_buf->priv->rb_chan = shmp(handle, handle->chan);
