@@ -28,7 +28,8 @@ int lib_ring_buffer_backend_create(struct lttng_ust_ring_buffer_backend *bufb,
 				   struct channel_backend *chan,
 				   int cpu,
 				   struct lttng_ust_shm_handle *handle,
-				   struct shm_object *shmobj)
+				   struct shm_object *shmobj,
+				   bool preallocate_backing)
 	__attribute__((visibility("hidden")));
 
 void channel_backend_unregister_notifiers(struct channel_backend *chanb)
@@ -42,7 +43,8 @@ int channel_backend_init(struct channel_backend *chanb,
 			 const struct lttng_ust_ring_buffer_config *config,
 			 size_t subbuf_size,
 			 size_t num_subbuf, struct lttng_ust_shm_handle *handle,
-			 const int *stream_fds)
+			 const int *stream_fds,
+			 bool preallocate_backing)
 	__attribute__((visibility("hidden")));
 
 void channel_backend_free(struct channel_backend *chanb,
@@ -105,16 +107,11 @@ extern void _lib_ring_buffer_write(struct lttng_ust_ring_buffer_backend *bufb,
  * mode).
  */
 static inline
-unsigned long subbuffer_id(const struct lttng_ust_ring_buffer_config *config,
+unsigned long subbuffer_id(const struct lttng_ust_ring_buffer_config *config __attribute__((unused)),
 			   unsigned long offset, unsigned long noref,
 			   unsigned long index)
 {
-	if (config->mode == RING_BUFFER_OVERWRITE)
-		return (offset << SB_ID_OFFSET_SHIFT)
-		       | (noref << SB_ID_NOREF_SHIFT)
-		       | index;
-	else
-		return index;
+	return (offset << SB_ID_OFFSET_SHIFT) | (noref << SB_ID_NOREF_SHIFT) | index;
 }
 
 /*
@@ -130,23 +127,17 @@ int subbuffer_id_compare_offset(
 }
 
 static inline
-unsigned long subbuffer_id_get_index(const struct lttng_ust_ring_buffer_config *config,
+unsigned long subbuffer_id_get_index(const struct lttng_ust_ring_buffer_config *config __attribute__((unused)),
 				     unsigned long id)
 {
-	if (config->mode == RING_BUFFER_OVERWRITE)
-		return id & SB_ID_INDEX_MASK;
-	else
-		return id;
+	return id & SB_ID_INDEX_MASK;
 }
 
 static inline
-unsigned long subbuffer_id_is_noref(const struct lttng_ust_ring_buffer_config *config,
+unsigned long subbuffer_id_is_noref(const struct lttng_ust_ring_buffer_config *config __attribute__((unused)),
 				    unsigned long id)
 {
-	if (config->mode == RING_BUFFER_OVERWRITE)
-		return !!(id & SB_ID_NOREF_MASK);
-	else
-		return 1;
+	return !!(id & SB_ID_NOREF_MASK);
 }
 
 /*
@@ -154,36 +145,32 @@ unsigned long subbuffer_id_is_noref(const struct lttng_ust_ring_buffer_config *c
  * needed.
  */
 static inline
-void subbuffer_id_set_noref(const struct lttng_ust_ring_buffer_config *config,
+void subbuffer_id_set_noref(const struct lttng_ust_ring_buffer_config *config __attribute__((unused)),
 			    unsigned long *id)
 {
-	if (config->mode == RING_BUFFER_OVERWRITE)
-		*id |= SB_ID_NOREF_MASK;
+	*id |= SB_ID_NOREF_MASK;
 }
 
 static inline
-void subbuffer_id_set_noref_offset(const struct lttng_ust_ring_buffer_config *config,
+void subbuffer_id_set_noref_offset(const struct lttng_ust_ring_buffer_config *config __attribute__((unused)),
 				   unsigned long *id, unsigned long offset)
 {
 	unsigned long tmp;
 
-	if (config->mode == RING_BUFFER_OVERWRITE) {
-		tmp = *id;
-		tmp &= ~SB_ID_OFFSET_MASK;
-		tmp |= offset << SB_ID_OFFSET_SHIFT;
-		tmp |= SB_ID_NOREF_MASK;
-		/* Volatile store, read concurrently by readers. */
-		CMM_ACCESS_ONCE(*id) = tmp;
-	}
+	tmp = *id;
+	tmp &= ~SB_ID_OFFSET_MASK;
+	tmp |= offset << SB_ID_OFFSET_SHIFT;
+	tmp |= SB_ID_NOREF_MASK;
+	/* Volatile store, read concurrently by readers. */
+	CMM_ACCESS_ONCE(*id) = tmp;
 }
 
 /* No volatile access, since already used locally */
 static inline
-void subbuffer_id_clear_noref(const struct lttng_ust_ring_buffer_config *config,
+void subbuffer_id_clear_noref(const struct lttng_ust_ring_buffer_config *config __attribute__((unused)),
 			      unsigned long *id)
 {
-	if (config->mode == RING_BUFFER_OVERWRITE)
-		*id &= ~SB_ID_NOREF_MASK;
+	*id &= ~SB_ID_NOREF_MASK;
 }
 
 /*
@@ -194,13 +181,10 @@ void subbuffer_id_clear_noref(const struct lttng_ust_ring_buffer_config *config,
  * -EPERM on failure.
  */
 static inline
-int subbuffer_id_check_index(const struct lttng_ust_ring_buffer_config *config,
+int subbuffer_id_check_index(const struct lttng_ust_ring_buffer_config *config __attribute__((unused)),
 			     unsigned long num_subbuf)
 {
-	if (config->mode == RING_BUFFER_OVERWRITE)
-		return (num_subbuf > (1UL << HALF_ULONG_BITS)) ? -EPERM : 0;
-	else
-		return 0;
+	return (num_subbuf > (1UL << HALF_ULONG_BITS)) ? -EPERM : 0;
 }
 
 static inline
@@ -230,8 +214,7 @@ int lib_ring_buffer_backend_get_pages(const struct lttng_ust_ring_buffer_config
 	if (caa_unlikely(!rpages))
 		return -1;
 	CHAN_WARN_ON(ctx_private->chan,
-		     config->mode == RING_BUFFER_OVERWRITE
-		     && subbuffer_id_is_noref(config, id));
+		     subbuffer_id_is_noref(config, id));
 	_backend_pages = shmp(handle, rpages->shmp);
 	if (caa_unlikely(!_backend_pages))
 		return -1;
@@ -437,18 +420,70 @@ unsigned long subbuffer_get_data_size(
 	return backend_pages->data_size;
 }
 
+/*
+ * Reader the packet counter of the sub-buffer at `idx`, given the current cold
+ * commit counter `current_commit_cold`.
+ *
+ * On error, return ~0ULL.
+ */
 static inline
-void subbuffer_inc_packet_count(
-		const struct lttng_ust_ring_buffer_config *config __attribute__((unused)),
-		struct lttng_ust_ring_buffer_backend *bufb,
-		unsigned long idx, struct lttng_ust_shm_handle *handle)
+uint64_t subbuffer_load_packet_count(
+		const struct lttng_ust_ring_buffer_channel *chan,
+		struct lttng_ust_ring_buffer *buf,
+		struct lttng_ust_shm_handle *handle,
+		unsigned long current_commit_cold,
+		unsigned long idx)
 {
 	struct lttng_ust_ring_buffer_backend_counts *counts;
+	size_t seq_idx;
 
-	counts = shmp_index(handle, bufb->buf_cnt, idx);
+	counts = shmp_index(handle, buf->backend.buf_cnt, idx);
+	if (!counts)
+		return ~0ULL;
+
+	seq_idx = (current_commit_cold >> chan->backend.subbuf_size_order) & 1;
+
+	return counts->seq_cnt[seq_idx];
+}
+
+/*
+ * Increment the packet count of sub-buffer at index `idx`, given the future
+ * value the cold commit counter `next_commit_cold`. Note that this only
+ * prepares the future slot and won't be observable by readers until the cold
+ * commit counter is moved to `next_commit_cold`.
+ *
+ * The sequence counter allows for atomic transaction that will be visible to
+ * reader once the cold commit counter is `next_commit_cold`.
+ *
+ * This works by preparing the value that will become the true value once the
+ * cold commit counter reached `next_commit_cold` by incrementing the value in
+ * the current slot and storing it in the next slot.
+ *
+ * Thus, the increment has the following properties:
+ *
+ *   - Mutually exclusive with respect to readers and writers
+ *
+ *   - Restartable by a fixup if the producer crashed before storing the
+ *     `next_commit_cold` into the cold commit counter
+ */
+static inline
+void subbuffer_inc_packet_count(
+		const struct lttng_ust_ring_buffer_channel *chan,
+		struct lttng_ust_ring_buffer *buf,
+		struct lttng_ust_shm_handle *handle,
+		unsigned long next_commit_cold,
+		unsigned long idx)
+{
+	struct lttng_ust_ring_buffer_backend_counts *counts;
+	size_t seq_idx;
+
+	counts = shmp_index(handle, buf->backend.buf_cnt, idx);
 	if (!counts)
 		return;
-	counts->seq_cnt++;
+
+	seq_idx = (next_commit_cold >> chan->backend.subbuf_size_order) & 1;
+
+	counts->seq_cnt[seq_idx] = counts->seq_cnt[1 - seq_idx] + 1;
 }
 
 /**
@@ -464,9 +499,6 @@ void lib_ring_buffer_clear_noref(const struct lttng_ust_ring_buffer_config *conf
 	unsigned long id, new_id;
 	struct lttng_ust_ring_buffer_backend_subbuffer *wsb;
 
-	if (config->mode != RING_BUFFER_OVERWRITE)
-		return;
-
 	/*
 	 * Performing a volatile access to read the sb_pages, because we want to
 	 * read a coherent version of the pointer and the associated noref flag.
@@ -508,9 +540,6 @@ void lib_ring_buffer_set_noref_offset(const struct lttng_ust_ring_buffer_config
 	struct lttng_ust_ring_buffer_backend_subbuffer *wsb;
 	struct lttng_ust_ring_buffer_channel *chan;
 
-	if (config->mode != RING_BUFFER_OVERWRITE)
-		return;
-
 	wsb = shmp_index(handle, bufb->buf_wsb, idx);
 	if (!wsb)
 		return;
@@ -549,45 +578,39 @@ int update_read_sb_index(const struct lttng_ust_ring_buffer_config *config,
 			 struct lttng_ust_shm_handle *handle)
 {
 	struct lttng_ust_ring_buffer_backend_subbuffer *wsb;
+	struct lttng_ust_ring_buffer_channel *chan;
 	unsigned long old_id, new_id;
 
 	wsb = shmp_index(handle, bufb->buf_wsb, consumed_idx);
 	if (caa_unlikely(!wsb))
 		return -EPERM;
 
-	if (config->mode == RING_BUFFER_OVERWRITE) {
-		struct lttng_ust_ring_buffer_channel *chan;
-
-		/*
-		 * Exchange the target writer subbuffer with our own unused
-		 * subbuffer. No need to use CMM_ACCESS_ONCE() here to read the
-		 * old_wpage, because the value read will be confirmed by the
-		 * following cmpxchg().
-		 */
-		old_id = wsb->id;
-		if (caa_unlikely(!subbuffer_id_is_noref(config, old_id)))
-			return -EAGAIN;
-		/*
-		 * Make sure the offset count we are expecting matches the one
-		 * indicated by the writer.
-		 */
-		if (caa_unlikely(!subbuffer_id_compare_offset(config, old_id,
-							  consumed_count)))
-			return -EAGAIN;
-		chan = shmp(handle, bufb->chan);
-		if (caa_unlikely(!chan))
-			return -EPERM;
-		CHAN_WARN_ON(chan, !subbuffer_id_is_noref(config, bufb->buf_rsb.id));
-		subbuffer_id_set_noref_offset(config, &bufb->buf_rsb.id,
-					      consumed_count);
-		new_id = uatomic_cmpxchg(&wsb->id, old_id, bufb->buf_rsb.id);
-		if (caa_unlikely(old_id != new_id))
-			return -EAGAIN;
-		bufb->buf_rsb.id = new_id;
-	} else {
-		/* No page exchange, use the writer page directly */
-		bufb->buf_rsb.id = wsb->id;
-	}
+	/*
+	 * Exchange the target writer subbuffer with our own unused
+	 * subbuffer. No need to use CMM_ACCESS_ONCE() here to read the
+	 * old_wpage, because the value read will be confirmed by the
+	 * following cmpxchg().
+	 */
+	old_id = wsb->id;
+	if (caa_unlikely(!subbuffer_id_is_noref(config, old_id)))
+		return -EAGAIN;
+	/*
+	 * Make sure the offset count we are expecting matches the one
+	 * indicated by the writer.
+	 */
+	if (caa_unlikely(!subbuffer_id_compare_offset(config, old_id,
+						  consumed_count)))
+		return -EAGAIN;
+	chan = shmp(handle, bufb->chan);
+	if (caa_unlikely(!chan))
+		return -EPERM;
+	CHAN_WARN_ON(chan, !subbuffer_id_is_noref(config, bufb->buf_rsb.id));
+	subbuffer_id_set_noref_offset(config, &bufb->buf_rsb.id,
+				      consumed_count);
+	new_id = uatomic_cmpxchg(&wsb->id, old_id, bufb->buf_rsb.id);
+	if (caa_unlikely(old_id != new_id))
+		return -EAGAIN;
+	bufb->buf_rsb.id = new_id;
 	return 0;
 }
 
@@ -662,45 +685,4 @@ void lib_ring_buffer_do_memset(char *dest, char c, unsigned long len)
 		dest[i] = c;
 }
 
-/* arch-agnostic implementation */
-
-static inline int lttng_ust_fls(unsigned int x)
-{
-	int r = 32;
-
-	if (!x)
-		return 0;
-	if (!(x & 0xFFFF0000U)) {
-		x <<= 16;
-		r -= 16;
-	}
-	if (!(x & 0xFF000000U)) {
-		x <<= 8;
-		r -= 8;
-	}
-	if (!(x & 0xF0000000U)) {
-		x <<= 4;
-		r -= 4;
-	}
-	if (!(x & 0xC0000000U)) {
-		x <<= 2;
-		r -= 2;
-	}
-	if (!(x & 0x80000000U)) {
-		/* No need to bit shift on last operation */
-		r -= 1;
-	}
-	return r;
-}
-
-static inline int get_count_order(unsigned int count)
-{
-	int order;
-
-	order = lttng_ust_fls(count) - 1;
-	if (count & (count - 1))
-		order++;
-	return order;
-}
-
 #endif /* _LTTNG_RING_BUFFER_BACKEND_INTERNAL_H */
