@@ -14,14 +14,48 @@
 #include "shm_internal.h"
 #include "vatomic.h"
 
-#define RB_BACKEND_PAGES_PADDING	16
+#define RB_BACKEND_PAGES_PADDING	128
 struct lttng_ust_ring_buffer_backend_pages {
 	unsigned long mmap_offset;	/* offset of the subbuffer in mmap */
 	union v_atomic records_commit;	/* current records committed count */
 	union v_atomic records_unread;	/* records to read */
 	unsigned long data_size;	/* Amount of data to read from subbuf */
 	DECLARE_SHMP(char, p);		/* Backing memory map */
-	char padding[RB_BACKEND_PAGES_PADDING];
+	union {
+		struct {
+			unsigned long begin_events_discarded;	/* Events discarded before starting sub-buffer. */
+			unsigned long end_events_discarded;	/*
+								 * Passing events discarded counter
+								 * read upon try_reserve and try_switch
+								 * that fills a subbuffer to check_deliver
+								 * so it can be written into the packet
+								 * header field.
+								 */
+			uint64_t timestamp_begin;		/*
+								 * timestamp_begin
+								 * per-subbuffer.
+								 */
+			uint64_t timestamp_end;			/*
+								 * timestamp_end per sub-buffer.
+								 * Time is sampled by the
+								 * switch_*_end() callbacks
+								 * which are the last space
+								 * reservation performed in the
+								 * sub-buffer before it can be
+								 * fully committed and
+								 * delivered. This time value is
+								 * then read by the deliver
+								 * callback, performed by the
+								 * last commit before the buffer
+								 * becomes readable.
+								 */
+			uint8_t allocated;			/*
+								 * Whether memory is allocated for
+								 * this subbuffer.
+								 */
+		};
+		char padding[RB_BACKEND_PAGES_PADDING];
+	};
 };
 
 struct lttng_ust_ring_buffer_backend_subbuffer {
@@ -33,10 +67,19 @@ struct lttng_ust_ring_buffer_backend_counts {
 	/*
 	 * Counter specific to the sub-buffer location within the ring buffer.
 	 * The actual sequence number of the packet within the entire ring
-	 * buffer can be derived from the formula nr_subbuffers * seq_cnt +
-	 * subbuf_idx.
+	 * buffer can be derived from the following formula:
+	 *
+	 *   nr_subbuffers * seq_cnt + subbuf_idx.
+	 *
+	 * This is a sequence counter with two slots.
+	 *
+	 * Readers always load from the current slot, while writers only mutate
+	 * the next slot. The current slot is derived from the cold commit
+	 * counter and the next slot is the other slot. See the corresponding
+	 * procedures `subbuffer_load_packet_count()` and
+	 * `subbuffer_inc_packet_count()` for details.
 	 */
-	uint64_t seq_cnt;               /* packet sequence number */
+	uint64_t seq_cnt[2];               /* packet sequence number */
 };
 
 /*
