@@ -22,6 +22,7 @@
 #include "shm.h"
 #include "common/align.h"
 #include "common/populate.h"
+#include "common/utils.h"
 
 /**
  * lib_ring_buffer_backend_allocate - allocate a channel buffer
@@ -37,7 +38,8 @@ int lib_ring_buffer_backend_allocate(const struct lttng_ust_ring_buffer_config *
 				     size_t size __attribute__((unused)), size_t num_subbuf,
 				     int extra_reader_sb,
 				     struct lttng_ust_shm_handle *handle,
-				     struct shm_object *shmobj)
+				     struct shm_object *shmobj,
+				     bool preallocate_backing)
 {
 	struct channel_backend *chanb;
 	unsigned long subbuf_size, mmap_offset = 0;
@@ -100,15 +102,15 @@ int lib_ring_buffer_backend_allocate(const struct lttng_ust_ring_buffer_config *
 		sb = shmp_index(handle, bufb->buf_wsb, i);
 		if (!sb)
 			goto free_array;
-		sb->id = subbuffer_id(config, 0, 1, i);
+		sb->id = subbuffer_id(config, ~0UL >> chanb->buf_size_order, 1, i);
 	}
 
 	/* Assign read-side subbuffer table */
 	if (extra_reader_sb)
-		bufb->buf_rsb.id = subbuffer_id(config, 0, 1,
+		bufb->buf_rsb.id = subbuffer_id(config, ~0UL, 1,
 						num_subbuf_alloc - 1);
 	else
-		bufb->buf_rsb.id = subbuffer_id(config, 0, 1, 0);
+		bufb->buf_rsb.id = subbuffer_id(config, ~0UL, 1, 0);
 
 	/* Allocate subbuffer packet counter table */
 	align_shm(shmobj, __alignof__(struct lttng_ust_ring_buffer_backend_counts));
@@ -139,6 +141,11 @@ int lib_ring_buffer_backend_allocate(const struct lttng_ust_ring_buffer_config *
 			pages->mmap_offset = mmap_offset;
 			mmap_offset += subbuf_size;
 		}
+		pages->allocated = preallocate_backing;
+		if (preallocate_backing) {
+			pages->timestamp_begin = chanb->start_timestamp;
+			pages->timestamp_end = chanb->start_timestamp;
+		}
 	}
 	return 0;
 
@@ -156,7 +163,7 @@ page_size_error:
 int lib_ring_buffer_backend_create(struct lttng_ust_ring_buffer_backend *bufb,
 				   struct channel_backend *chanb, int cpu,
 				   struct lttng_ust_shm_handle *handle,
-				   struct shm_object *shmobj)
+				   struct shm_object *shmobj, bool preallocate_backing)
 {
 	const struct lttng_ust_ring_buffer_config *config = &chanb->config;
 
@@ -166,7 +173,7 @@ int lib_ring_buffer_backend_create(struct lttng_ust_ring_buffer_backend *bufb,
 	return lib_ring_buffer_backend_allocate(config, bufb, chanb->buf_size,
 						chanb->num_subbuf,
 						chanb->extra_reader_sb,
-						handle, shmobj);
+						handle, shmobj, preallocate_backing);
 }
 
 void lib_ring_buffer_backend_reset(struct lttng_ust_ring_buffer_backend *bufb,
@@ -248,6 +255,8 @@ void channel_backend_reset(struct channel_backend *chanb)
  * @num_subbuf: number of sub-buffers (power of 2)
  * @lttng_ust_shm_handle: shared memory handle
  * @stream_fds: stream file descriptors.
+ * @preallocate_backing: preallocate the channel buffers if true,
+ *                       otherwise allocate them on demand.
  *
  * Returns channel pointer if successful, %NULL otherwise.
  *
@@ -262,7 +271,7 @@ int channel_backend_init(struct channel_backend *chanb,
 			 const struct lttng_ust_ring_buffer_config *config,
 			 size_t subbuf_size, size_t num_subbuf,
 			 struct lttng_ust_shm_handle *handle,
-			 const int *stream_fds)
+			 const int *stream_fds, bool preallocate_backing)
 {
 	struct lttng_ust_ring_buffer_channel *chan = caa_container_of(chanb,
 			struct lttng_ust_ring_buffer_channel, backend);
@@ -291,23 +300,26 @@ int channel_backend_init(struct channel_backend *chanb,
 	if (!num_subbuf || (num_subbuf & (num_subbuf - 1)))
 		return -EINVAL;
 	/*
-	 * Overwrite mode buffers require at least 2 subbuffers per
-	 * buffer.
+	 * Require at least 2 subbuffers per buffer.
 	 */
-	if (config->mode == RING_BUFFER_OVERWRITE && num_subbuf < 2)
+	if (num_subbuf < 2)
 		return -EINVAL;
 
 	ret = subbuffer_id_check_index(config, num_subbuf);
 	if (ret)
 		return ret;
 
+	chanb->start_timestamp = config->cb.ring_buffer_clock_read(chan);
 	chanb->buf_size = num_subbuf * subbuf_size;
 	chanb->subbuf_size = subbuf_size;
-	chanb->buf_size_order = get_count_order(chanb->buf_size);
-	chanb->subbuf_size_order = get_count_order(subbuf_size);
-	chanb->num_subbuf_order = get_count_order(num_subbuf);
-	chanb->extra_reader_sb =
-			(config->mode == RING_BUFFER_OVERWRITE) ? 1 : 0;
+	chanb->buf_size_order = lttng_ust_get_count_order_ulong(chanb->buf_size);
+	chanb->subbuf_size_order = lttng_ust_get_count_order_ulong(subbuf_size);
+	chanb->num_subbuf_order = lttng_ust_get_count_order_ulong(num_subbuf);
+	/*
+	 * Allocate an extra subbuffer for the reader to allow punching
+	 * holes (reclaim pages) from any type of ring buffer.
+	 */
+	chanb->extra_reader_sb = 1;
 	chanb->num_subbuf = num_subbuf;
 	strncpy(chanb->name, name, NAME_MAX);
 	chanb->name[NAME_MAX - 1] = '\0';
@@ -320,9 +332,6 @@ int channel_backend_init(struct channel_backend *chanb,
 	shmsize += sizeof(struct commit_counters_hot) * num_subbuf;
 	shmsize += lttng_ust_offset_align(shmsize, __alignof__(struct commit_counters_cold));
 	shmsize += sizeof(struct commit_counters_cold) * num_subbuf;
-	/* Sampled timestamp end */
-	shmsize += lttng_ust_offset_align(shmsize, __alignof__(uint64_t));
-	shmsize += sizeof(uint64_t) * num_subbuf;
 
 	/* Per-cpu buffer size: backend */
 	/* num_subbuf + 1 is the worse case */
@@ -344,11 +353,12 @@ int channel_backend_init(struct channel_backend *chanb,
 		 * We need to allocate for all possible cpus.
 		 */
 		for_each_possible_cpu(i) {
+			bool populate = lttng_ust_map_populate_cpu_is_enabled(i);
 			struct shm_object *shmobj;
 
 			shmobj = shm_object_table_alloc(handle->table, shmsize,
-					SHM_OBJECT_SHM, stream_fds[i], i,
-					lttng_ust_map_populate_cpu_is_enabled(i));
+					SHM_OBJECT_SHM, stream_fds[i], i, populate,
+					preallocate_backing);
 			if (!shmobj)
 				goto end;
 			align_shm(shmobj, __alignof__(struct lttng_ust_ring_buffer));
@@ -358,17 +368,19 @@ int channel_backend_init(struct channel_backend *chanb,
 				goto end;
 			set_shmp(buf->self, chanb->buf[i].shmp._ref);
 			ret = lib_ring_buffer_create(buf, chanb, i,
-					handle, shmobj);
+					handle, shmobj, preallocate_backing);
 			if (ret)
 				goto free_bufs;	/* cpu hotplug locked */
 		}
 	} else {
+		bool populate = lttng_ust_map_populate_is_enabled();
 		struct shm_object *shmobj;
 		struct lttng_ust_ring_buffer *buf;
 
 		shmobj = shm_object_table_alloc(handle->table, shmsize,
 					SHM_OBJECT_SHM, stream_fds[0], -1,
-					lttng_ust_map_populate_is_enabled());
+					populate,
+					preallocate_backing);
 		if (!shmobj)
 			goto end;
 		align_shm(shmobj, __alignof__(struct lttng_ust_ring_buffer));
@@ -378,11 +390,10 @@ int channel_backend_init(struct channel_backend *chanb,
 			goto end;
 		set_shmp(buf->self, chanb->buf[0].shmp._ref);
 		ret = lib_ring_buffer_create(buf, chanb, -1,
-					handle, shmobj);
+					handle, shmobj, preallocate_backing);
 		if (ret)
 			goto free_bufs;
 	}
-	chanb->start_timestamp = config->cb.ring_buffer_clock_read(chan);
 
 	return 0;
 
@@ -444,8 +455,7 @@ size_t lib_ring_buffer_read(struct lttng_ust_ring_buffer_backend *bufb, size_t o
 	 * subbuffers.
 	 */
 	CHAN_WARN_ON(chanb, offset >= chanb->buf_size);
-	CHAN_WARN_ON(chanb, config->mode == RING_BUFFER_OVERWRITE
-		     && subbuffer_id_is_noref(config, id));
+	CHAN_WARN_ON(chanb, subbuffer_id_is_noref(config, id));
 	backend_pages = shmp(handle, rpages->shmp);
 	if (!backend_pages)
 		return 0;
@@ -496,8 +506,7 @@ int lib_ring_buffer_read_cstr(struct lttng_ust_ring_buffer_backend *bufb, size_t
 	 * subbuffers.
 	 */
 	CHAN_WARN_ON(chanb, offset >= chanb->buf_size);
-	CHAN_WARN_ON(chanb, config->mode == RING_BUFFER_OVERWRITE
-		     && subbuffer_id_is_noref(config, id));
+	CHAN_WARN_ON(chanb, subbuffer_id_is_noref(config, id));
 	backend_pages = shmp(handle, rpages->shmp);
 	if (!backend_pages)
 		return -EINVAL;
@@ -512,6 +521,34 @@ int lib_ring_buffer_read_cstr(struct lttng_ust_ring_buffer_backend *bufb, size_t
 	return offset - orig_offset;
 }
 
+/**
+ * lib_ring_buffer_read_backend_pages - get backend pages of a buffer location
+ * @bufb : buffer backend
+ */
+struct lttng_ust_ring_buffer_backend_pages *
+	lib_ring_buffer_read_backend_pages(struct lttng_ust_ring_buffer_backend *bufb,
+					   struct lttng_ust_shm_handle *handle)
+{
+	struct lttng_ust_ring_buffer_backend_pages_shmp *rpages;
+	struct lttng_ust_ring_buffer_backend_pages *backend_pages;
+	struct channel_backend *chanb;
+	const struct lttng_ust_ring_buffer_config *config;
+	unsigned long sb_bindex, id;
+
+	chanb = &shmp(handle, bufb->chan)->backend;
+	if (!chanb)
+		return NULL;
+	config = &chanb->config;
+	id = bufb->buf_rsb.id;
+	sb_bindex = subbuffer_id_get_index(config, id);
+	rpages = shmp_index(handle, bufb->array, sb_bindex);
+	if (!rpages)
+		return NULL;
+	CHAN_WARN_ON(chanb, subbuffer_id_is_noref(config, id));
+	backend_pages = shmp(handle, rpages->shmp);
+	return backend_pages;
+}
+
 /**
  * lib_ring_buffer_read_offset_address - get address of a buffer location
  * @bufb : buffer backend
@@ -526,29 +563,72 @@ int lib_ring_buffer_read_cstr(struct lttng_ust_ring_buffer_backend *bufb, size_t
 void *lib_ring_buffer_read_offset_address(struct lttng_ust_ring_buffer_backend *bufb,
 					  size_t offset,
 					  struct lttng_ust_shm_handle *handle)
+{
+	struct lttng_ust_ring_buffer_backend_pages *backend_pages;
+	struct channel_backend *chanb;
+
+	chanb = &shmp(handle, bufb->chan)->backend;
+	if (!chanb)
+		return NULL;
+	backend_pages = lib_ring_buffer_read_backend_pages(bufb, handle);
+	if (!backend_pages)
+		return NULL;
+	return shmp_index(handle, backend_pages->p, offset & (chanb->subbuf_size - 1));
+}
+
+/**
+ * lib_ring_buffer_index_backend_pages - get the backend pages of a subbuffer index
+ * @bufb : buffer backend
+ * @subbuf_idx: subbuffer index.
+ */
+struct lttng_ust_ring_buffer_backend_pages *lib_ring_buffer_index_backend_pages(
+		struct lttng_ust_ring_buffer_backend *bufb, size_t sbidx,
+		struct lttng_ust_shm_handle *handle)
 {
 	struct lttng_ust_ring_buffer_backend_pages_shmp *rpages;
 	struct lttng_ust_ring_buffer_backend_pages *backend_pages;
 	struct channel_backend *chanb;
 	const struct lttng_ust_ring_buffer_config *config;
 	unsigned long sb_bindex, id;
+	struct lttng_ust_ring_buffer_backend_subbuffer *sb;
 
 	chanb = &shmp(handle, bufb->chan)->backend;
 	if (!chanb)
 		return NULL;
 	config = &chanb->config;
-	offset &= chanb->buf_size - 1;
-	id = bufb->buf_rsb.id;
+	sb = shmp_index(handle, bufb->buf_wsb, sbidx);
+	if (!sb)
+		return NULL;
+	id = sb->id;
 	sb_bindex = subbuffer_id_get_index(config, id);
 	rpages = shmp_index(handle, bufb->array, sb_bindex);
 	if (!rpages)
 		return NULL;
-	CHAN_WARN_ON(chanb, config->mode == RING_BUFFER_OVERWRITE
-		     && subbuffer_id_is_noref(config, id));
+	CHAN_WARN_ON(chanb, subbuffer_id_is_noref(config, id));
 	backend_pages = shmp(handle, rpages->shmp);
 	if (!backend_pages)
 		return NULL;
-	return shmp_index(handle, backend_pages->p, offset & (chanb->subbuf_size - 1));
+	return backend_pages;
+}
+
+/**
+ * lib_ring_buffer_offset_backend_pages - get the backend pages of a location within the buffer
+ * @bufb : buffer backend
+ * @offset : offset within the buffer.
+ */
+struct lttng_ust_ring_buffer_backend_pages *lib_ring_buffer_offset_backend_pages(
+		struct lttng_ust_ring_buffer_backend *bufb, size_t offset,
+		struct lttng_ust_shm_handle *handle)
+{
+	struct channel_backend *chanb;
+	unsigned long sbidx;
+
+	chanb = &shmp(handle, bufb->chan)->backend;
+	if (!chanb)
+		return NULL;
+	offset &= chanb->buf_size - 1;
+	sbidx = offset >> chanb->subbuf_size_order;
+	return lib_ring_buffer_index_backend_pages(bufb, sbidx, handle);
 }
 
 /**
@@ -565,31 +645,14 @@ void *lib_ring_buffer_offset_address(struct lttng_ust_ring_buffer_backend *bufb,
 				     size_t offset,
 				     struct lttng_ust_shm_handle *handle)
 {
-	size_t sbidx;
-	struct lttng_ust_ring_buffer_backend_pages_shmp *rpages;
-	struct lttng_ust_ring_buffer_backend_pages *backend_pages;
+	struct lttng_ust_ring_buffer_backend_pages *backend_pages =
+		lib_ring_buffer_offset_backend_pages(bufb, offset, handle);
 	struct channel_backend *chanb;
-	const struct lttng_ust_ring_buffer_config *config;
-	unsigned long sb_bindex, id;
-	struct lttng_ust_ring_buffer_backend_subbuffer *sb;
 
 	chanb = &shmp(handle, bufb->chan)->backend;
 	if (!chanb)
 		return NULL;
-	config = &chanb->config;
-	offset &= chanb->buf_size - 1;
-	sbidx = offset >> chanb->subbuf_size_order;
-	sb = shmp_index(handle, bufb->buf_wsb, sbidx);
-	if (!sb)
-		return NULL;
-	id = sb->id;
-	sb_bindex = subbuffer_id_get_index(config, id);
-	rpages = shmp_index(handle, bufb->array, sb_bindex);
-	if (!rpages)
-		return NULL;
-	CHAN_WARN_ON(chanb, config->mode == RING_BUFFER_OVERWRITE
-		     && subbuffer_id_is_noref(config, id));
-	backend_pages = shmp(handle, rpages->shmp);
+	backend_pages = lib_ring_buffer_offset_backend_pages(bufb, offset, handle);
 	if (!backend_pages)
 		return NULL;
 	return shmp_index(handle, backend_pages->p, offset & (chanb->subbuf_size - 1));
