@@ -64,6 +64,8 @@
 #include "rb-init.h"
 #include "common/compat/errno.h"	/* For ENODATA */
 #include "common/populate.h"
+#include "common/testpoint.h"
+#include "common/utils.h"
 
 /* Print DBG() messages about events lost only every 1048576 hits */
 #define DBG_PRINT_NR_LOST	(1UL << 20)
@@ -185,7 +187,6 @@ void lib_ring_buffer_reset(struct lttng_ust_ring_buffer *buf,
 	for (i = 0; i < chan->backend.num_subbuf; i++) {
 		struct commit_counters_hot *cc_hot;
 		struct commit_counters_cold *cc_cold;
-		uint64_t *ts_end;
 
 		cc_hot = shmp_index(handle, buf->commit_hot, i);
 		if (!cc_hot)
@@ -193,18 +194,14 @@ void lib_ring_buffer_reset(struct lttng_ust_ring_buffer *buf,
 		cc_cold = shmp_index(handle, buf->commit_cold, i);
 		if (!cc_cold)
 			return;
-		ts_end = shmp_index(handle, buf->ts_end, i);
-		if (!ts_end)
-			return;
 		v_set(config, &cc_hot->cc, 0);
 		v_set(config, &cc_hot->seq, 0);
 		v_set(config, &cc_cold->cc_sb, 0);
-		cc_cold->end_events_discarded = 0;
-		*ts_end = 0;
 	}
 	uatomic_set(&buf->consumed, 0);
 	uatomic_set(&buf->record_disabled, 0);
-	v_set(config, &buf->last_timestamp, 0);
+	uatomic_set(&buf->use_creation_timestamp, 1);
+	save_last_timestamp(config, chan, buf, chan->backend.start_timestamp);
 	lib_ring_buffer_backend_reset(&buf->backend, handle);
 	/* Don't reset number of active readers */
 	v_set(config, &buf->records_lost_full, 0);
@@ -332,17 +329,12 @@ void init_crash_abi(const struct lttng_ust_ring_buffer_config *config,
 int lib_ring_buffer_create(struct lttng_ust_ring_buffer *buf,
 			   struct channel_backend *chanb, int cpu,
 			   struct lttng_ust_shm_handle *handle,
-			   struct shm_object *shmobj)
+			   struct shm_object *shmobj, bool preallocate_backing)
 {
 	const struct lttng_ust_ring_buffer_config *config = &chanb->config;
 	struct lttng_ust_ring_buffer_channel *chan = caa_container_of(chanb,
 			struct lttng_ust_ring_buffer_channel, backend);
-	struct lttng_ust_ring_buffer_backend_subbuffer *wsb;
-	struct lttng_ust_ring_buffer_channel *shmp_chan;
-	struct commit_counters_hot *cc_hot;
 	void *priv = channel_get_private_config(chan);
-	size_t subbuf_header_size;
-	uint64_t timestamp;
 	int ret;
 
 	/* Test for cpu hotplug */
@@ -366,48 +358,19 @@ int lib_ring_buffer_create(struct lttng_ust_ring_buffer *buf,
 		goto free_commit;
 	}
 
-	align_shm(shmobj, __alignof__(uint64_t));
-	set_shmp(buf->ts_end,
-		 zalloc_shm(shmobj,
-			sizeof(uint64_t) * chan->backend.num_subbuf));
-	if (!shmp(handle, buf->ts_end)) {
-		ret = -ENOMEM;
-		goto free_commit_cold;
-	}
-
-
 	ret = lib_ring_buffer_backend_create(&buf->backend, &chan->backend,
-			cpu, handle, shmobj);
+			cpu, handle, shmobj, preallocate_backing);
 	if (ret) {
-		goto free_init;
+		goto free_commit_cold;
 	}
 
 	/*
 	 * Write the subbuffer header for first subbuffer so we know the total
 	 * duration of data gathering.
 	 */
-	subbuf_header_size = config->cb.subbuffer_header_size();
-	v_set(config, &buf->offset, subbuf_header_size);
-	wsb = shmp_index(handle, buf->backend.buf_wsb, 0);
-	if (!wsb) {
-		ret = -EPERM;
-		goto free_chanbuf;
-	}
-	subbuffer_id_clear_noref(config, &wsb->id);
-	shmp_chan = shmp(handle, buf->backend.chan);
-	if (!shmp_chan) {
-		ret = -EPERM;
-		goto free_chanbuf;
-	}
-	timestamp = config->cb.ring_buffer_clock_read(shmp_chan);
-	config->cb.buffer_begin(buf, timestamp, 0, handle);
-	cc_hot = shmp_index(handle, buf->commit_hot, 0);
-	if (!cc_hot) {
-		ret = -EPERM;
-		goto free_chanbuf;
-	}
-	v_add(config, subbuf_header_size, &cc_hot->cc);
-	v_add(config, subbuf_header_size, &cc_hot->seq);
+	v_set(config, &buf->offset, 0);
+	uatomic_set(&buf->use_creation_timestamp, 1);
+	save_last_timestamp(config, chan, buf, chanb->start_timestamp);
 
 	if (config->cb.buffer_create) {
 		ret = config->cb.buffer_create(buf, priv, cpu, chanb->name, handle);
@@ -421,8 +384,6 @@ int lib_ring_buffer_create(struct lttng_ust_ring_buffer *buf,
 	return 0;
 
 	/* Error handling */
-free_init:
-	/* ts_end will be freed by shm teardown */
 free_commit_cold:
 	/* commit_cold will be freed by shm teardown */
 free_commit:
@@ -461,7 +422,7 @@ void lib_ring_buffer_channel_switch_timer(int sig __attribute__((unused)),
 				goto end;
 			if (uatomic_read(&buf->active_readers))
 				lib_ring_buffer_switch_slow(buf, SWITCH_ACTIVE,
-					chan->handle);
+							NULL, chan->handle);
 		}
 	} else {
 		struct lttng_ust_ring_buffer *buf =
@@ -471,7 +432,7 @@ void lib_ring_buffer_channel_switch_timer(int sig __attribute__((unused)),
 			goto end;
 		if (uatomic_read(&buf->active_readers))
 			lib_ring_buffer_switch_slow(buf, SWITCH_ACTIVE,
-				chan->handle);
+						NULL, chan->handle);
 	}
 end:
 	pthread_mutex_unlock(&wakeup_fd_mutex);
@@ -524,7 +485,6 @@ int lib_ring_buffer_poll_deliver(const struct lttng_ust_ring_buffer_config *conf
 	return 1;
 }
 
-static
 void lib_ring_buffer_wakeup(struct lttng_ust_ring_buffer *buf,
 		struct lttng_ust_shm_handle *handle)
 {
@@ -939,6 +899,28 @@ static void channel_free(struct lttng_ust_ring_buffer_channel *chan,
 	shm_object_table_destroy(handle->table, consumer);
 	free(handle);
 }
+#if (CAA_BITS_PER_LONG == 32)
+static unsigned int get_activity_timestamp_bits(const struct lttng_ust_ring_buffer_config *config,
+						struct lttng_ust_ring_buffer_channel *chan)
+{
+	uint64_t freq = 0;
+
+	if (config->cb.ring_buffer_clock_freq)
+		freq = config->cb.ring_buffer_clock_freq(chan);
+	if (!freq)
+		return 0;
+	/*
+	 * The activity timestamp accuracy is in the order of 250ms (1s / 4).
+	 */
+	return lttng_ust_get_count_order_u64(freq >> 2);
+}
+#else
+static unsigned int get_activity_timestamp_bits(const struct lttng_ust_ring_buffer_config *config __attribute__((unused)),
+						struct lttng_ust_ring_buffer_channel *chan __attribute__((unused)))
+{
+	return 0;
+}
+#endif
 
 /**
  * channel_create - Create channel.
@@ -959,6 +941,8 @@ static void channel_free(struct lttng_ust_ring_buffer_channel *chan,
  * @read_timer_interval: Time interval (in us) to wake up pending readers.
  * @stream_fds: array of stream file descriptors.
  * @nr_stream_fds: number of file descriptors in array.
+ * @preallocate_backing: whether to preallocate the backing memory for the
+ *                       ring buffer or do so lazily on access.
  *
  * Holds cpu hotplug.
  * Returns NULL on failure.
@@ -973,7 +957,9 @@ struct lttng_ust_shm_handle *channel_create(const struct lttng_ust_ring_buffer_c
 		   size_t num_subbuf, unsigned int switch_timer_interval,
 		   unsigned int read_timer_interval,
 		   const int *stream_fds, int nr_stream_fds,
-		   int64_t blocking_timeout)
+		   int64_t blocking_timeout,
+		   uint32_t owner_id,
+		   bool preallocate_backing)
 {
 	int ret;
 	size_t shmsize, chansize;
@@ -1029,7 +1015,7 @@ struct lttng_ust_shm_handle *channel_create(const struct lttng_ust_ring_buffer_c
 
 	/* Allocate normal memory for channel (not shared) */
 	shmobj = shm_object_table_alloc(handle->table, shmsize, SHM_OBJECT_MEM,
-			-1, -1, populate);
+			-1, -1, populate, preallocate_backing);
 	if (!shmobj)
 		goto error_append;
 	/* struct lttng_ust_ring_buffer_channel is at object 0, offset 0 (hardcoded) */
@@ -1059,12 +1045,14 @@ struct lttng_ust_shm_handle *channel_create(const struct lttng_ust_ring_buffer_c
 	}
 
 	chan->u.s.blocking_timeout_ms = (int32_t) blocking_timeout_ms;
+	chan->u.s.owner_id = owner_id;
+	chan->u.s.activity_timestamp_bits = get_activity_timestamp_bits(config, chan);
 
 	channel_set_private(chan, priv);
 
 	ret = channel_backend_init(&chan->backend, name, config,
 				   subbuf_size, num_subbuf, handle,
-				   stream_fds);
+				   stream_fds, preallocate_backing);
 	if (ret)
 		goto error_backend_init;
 
@@ -1417,6 +1405,104 @@ void lib_ring_buffer_move_consumer(struct lttng_ust_ring_buffer *buf,
 					   consumed_new);
 }
 
+/**
+ * lib_ring_buffer_try_exchange_subbuf - exchange subbuffer at pos with reader subbuffer
+ * @buf: ring buffer
+ * @pos: produce position to exchange
+ *
+ * Return -ENOENT if the subbuffer is being used for writing or there is
+ * no subbuffer matching the @pos producer position in the buffer, else
+ * return 0 on success. Return -EIO on shared memory access error.
+ */
+int lib_ring_buffer_try_exchange_subbuf(struct lttng_ust_ring_buffer *buf,
+	unsigned long pos, struct lttng_ust_shm_handle *handle)
+{
+	const struct lttng_ust_ring_buffer_config *config;
+	struct lttng_ust_ring_buffer_channel *chan;
+	unsigned long idx;
+	int ret;
+
+	chan = shmp(handle, buf->backend.chan);
+	if (!chan)
+		return -EIO;
+	config = &chan->backend.config;
+
+	/*
+	 * Move the consumer forward one subbuffer ahead of pos if we
+	 * are about to exchange a yet unconsumed subbuffer.
+	 */
+	lib_ring_buffer_move_consumer(buf, subbuf_align(pos, chan), handle);
+
+	idx = subbuf_index(pos, chan);
+	ret = update_read_sb_index(config, &buf->backend, &chan->backend,
+				   idx, buf_trunc_val(pos, chan),
+				   handle);
+	switch (ret) {
+	case -EPERM:
+		return -EIO;
+	case -EAGAIN:
+		return -ENOENT;
+	}
+	subbuffer_id_clear_noref(config, &buf->backend.buf_rsb.id);
+	return 0;
+}
+
+#ifdef __linux__
+/**
+ * lib_ring_buffer_reclaim_reader_subbuf - Reclaim memory (MADV_REMOVE) backing reader subbuffer.
+ * @buf: ring buffer
+ *
+ * Return -ENOENT if the subbuffer is being used for writing or there is
+ * no subbuffer matching the @pos producer position in the buffer, else
+ * return 0 on success. Return -EIO on shared memory access error.
+ */
+int lib_ring_buffer_reclaim_reader_subbuf(struct lttng_ust_ring_buffer *buf,
+	struct lttng_ust_shm_handle *handle)
+{
+	const struct lttng_ust_ring_buffer_config *config;
+	struct lttng_ust_ring_buffer_channel *chan;
+	struct lttng_ust_ring_buffer_backend_pages_shmp *rpages;
+	struct lttng_ust_ring_buffer_backend_pages *backend_pages;
+	unsigned long id, sb_bindex;
+	void *addr;
+	int ret;
+
+	chan = shmp(handle, buf->backend.chan);
+	if (!chan)
+		return -EIO;
+	config = &chan->backend.config;
+	id = buf->backend.buf_rsb.id;
+	sb_bindex = subbuffer_id_get_index(config, id);
+	rpages = shmp_index(handle, buf->backend.array, sb_bindex);
+	if (!rpages)
+		return -EIO;
+	backend_pages = shmp(handle, rpages->shmp);
+	if (!backend_pages)
+		return -EIO;
+	if (!backend_pages->allocated)
+		return -ENOMEM;
+	addr = shmp_index(handle, backend_pages->p, 0);
+	if (!addr)
+		return -EIO;
+	ret = madvise(addr, chan->backend.subbuf_size, MADV_REMOVE);
+	if (ret)
+		return -errno;
+	backend_pages->allocated = false;
+	backend_pages->timestamp_begin = 0;
+	backend_pages->timestamp_end = 0;
+	return 0;
+}
+#else
+/*
+ * MADV_REMOVE is linux specific.
+ */
+int lib_ring_buffer_reclaim_reader_subbuf(struct lttng_ust_ring_buffer *buf,
+	struct lttng_ust_shm_handle *handle)
+{
+	return -EOPNOTSUPP;
+}
+#endif
+
 /**
  * lib_ring_buffer_get_subbuf - get exclusive access to subbuffer for reading
  * @buf: ring buffer
@@ -1632,8 +1718,7 @@ void lib_ring_buffer_put_subbuf(struct lttng_ust_ring_buffer *buf,
 	v_add(config, v_read(config, &backend_pages->records_unread),
 			&bufb->records_read);
 	v_set(config, &backend_pages->records_unread, 0);
-	CHAN_WARN_ON(chan, config->mode == RING_BUFFER_OVERWRITE
-		     && subbuffer_id_is_noref(config, bufb->buf_rsb.id));
+	CHAN_WARN_ON(chan, subbuffer_id_is_noref(config, bufb->buf_rsb.id));
 	subbuffer_id_set_noref(config, &bufb->buf_rsb.id);
 
 	/*
@@ -1775,7 +1860,12 @@ void lib_ring_buffer_switch_old_start(struct lttng_ust_ring_buffer *buf,
 	unsigned long commit_count;
 	struct commit_counters_hot *cc_hot;
 
-	config->cb.buffer_begin(buf, ctx->priv->timestamp, oldidx, handle);
+	config->cb.buffer_begin(buf,
+				ctx->priv->timestamp,
+				ctx->priv->records_lost_full +
+				ctx->priv->records_lost_wrap +
+				ctx->priv->records_lost_big,
+				oldidx, handle);
 
 	/*
 	 * Order all writes to buffer before the commit count update that will
@@ -1787,6 +1877,13 @@ void lib_ring_buffer_switch_old_start(struct lttng_ust_ring_buffer *buf,
 		return;
 	v_add(config, config->cb.subbuffer_header_size(),
 	      &cc_hot->cc);
+
+	/*
+	 * Observable side-effect: The old sub-buffer packet header begin has been
+	 * commited.
+	 */
+	TESTPOINT("lib_ring_buffer_switch_old_start_after_commit");
+
 	commit_count = v_read(config, &cc_hot->cc);
 	/* Check if the written buffer has to be delivered */
 	lib_ring_buffer_check_deliver(config, buf, chan, offsets->old,
@@ -1814,42 +1911,50 @@ void lib_ring_buffer_switch_old_end(struct lttng_ust_ring_buffer *buf,
 	const struct lttng_ust_ring_buffer_config *config = &chan->backend.config;
 	unsigned long oldidx = subbuf_index(offsets->old - 1, chan);
 	unsigned long commit_count, padding_size, data_size;
-	struct commit_counters_cold *cc_cold;
+	struct lttng_ust_ring_buffer_backend_pages *backend_pages =
+			lib_ring_buffer_offset_backend_pages(&buf->backend,
+				offsets->old - 1, handle);
 	struct commit_counters_hot *cc_hot;
-	uint64_t *ts_end;
 
 	data_size = subbuf_offset(offsets->old - 1, chan) + 1;
 	padding_size = chan->backend.subbuf_size - data_size;
 	subbuffer_set_data_size(config, &buf->backend, oldidx, data_size,
 				handle);
 
-	ts_end = shmp_index(handle, buf->ts_end, oldidx);
-	cc_cold = shmp_index(handle, buf->commit_cold, oldidx);
-	if (!ts_end || !cc_cold)
+	if (!backend_pages)
 		return;
 	/*
 	 * This is the last space reservation in that sub-buffer before
 	 * it gets delivered. This provides exclusive access to write to
-	 * this sub-buffer's ts_end. There are also no concurrent
-	 * readers of that ts_end because delivery of that sub-buffer is
-	 * postponed until the commit counter is incremented for the
-	 * current space reservation.
+	 * this sub-buffer's timestamp_end. There are also no concurrent
+	 * readers of that timestamp_end because delivery of that
+	 * sub-buffer is postponed until the commit counter is
+	 * incremented for the current space reservation.
 	 */
-	*ts_end = ctx->priv->timestamp;
-	cc_cold->end_events_discarded =
+	backend_pages->allocated = true;
+	backend_pages->timestamp_end = ctx->priv->timestamp;
+	backend_pages->end_events_discarded =
 		lib_ring_buffer_get_records_lost_full(config, ctx) +
 		lib_ring_buffer_get_records_lost_wrap(config, ctx) +
 		lib_ring_buffer_get_records_lost_big(config, ctx);
 
 	/*
-	 * Order all writes to buffer and store to ts_end before the commit
-	 * count update that will determine that the subbuffer is full.
+	 * Order all writes to buffer and store to timestamp_end before
+	 * the commit count update that will determine that the
+	 * subbuffer is full.
 	 */
 	cmm_smp_wmb();
 	cc_hot = shmp_index(handle, buf->commit_hot, oldidx);
 	if (!cc_hot)
 		return;
 	v_add(config, padding_size, &cc_hot->cc);
+
+	/*
+	 * Observable side-effect: The old sub-buffer packet header end has been
+	 * commited.
+	 */
+	TESTPOINT("lib_ring_buffer_switch_old_end_after_commit");
+
 	commit_count = v_read(config, &cc_hot->cc);
 	lib_ring_buffer_check_deliver(config, buf, chan, offsets->old - 1,
 				      commit_count, oldidx, handle, ctx);
@@ -1877,7 +1982,12 @@ void lib_ring_buffer_switch_new_start(struct lttng_ust_ring_buffer *buf,
 	unsigned long commit_count;
 	struct commit_counters_hot *cc_hot;
 
-	config->cb.buffer_begin(buf, ctx->priv->timestamp, beginidx, handle);
+	config->cb.buffer_begin(buf,
+				ctx->priv->timestamp,
+				ctx->priv->records_lost_full +
+				ctx->priv->records_lost_big +
+				ctx->priv->records_lost_wrap,
+				beginidx, handle);
 
 	/*
 	 * Order all writes to buffer before the commit count update that will
@@ -1888,6 +1998,13 @@ void lib_ring_buffer_switch_new_start(struct lttng_ust_ring_buffer *buf,
 	if (!cc_hot)
 		return;
 	v_add(config, config->cb.subbuffer_header_size(), &cc_hot->cc);
+
+	/*
+	 * Observable side-effect: The new sub-buffer packet header begin has been
+	 * commited.
+	 */
+	TESTPOINT("lib_ring_buffer_switch_new_start_after_commit");
+
 	commit_count = v_read(config, &cc_hot->cc);
 	/* Check if the written buffer has to be delivered */
 	lib_ring_buffer_check_deliver(config, buf, chan, offsets->begin,
@@ -1913,28 +2030,27 @@ void lib_ring_buffer_switch_new_end(struct lttng_ust_ring_buffer *buf,
 				    struct lttng_ust_shm_handle *handle)
 {
 	const struct lttng_ust_ring_buffer_config *config = &chan->backend.config;
-	struct commit_counters_cold *cc_cold;
+	struct lttng_ust_ring_buffer_backend_pages *backend_pages =
+			lib_ring_buffer_offset_backend_pages(&buf->backend,
+				offsets->end - 1, handle);
 	unsigned long endidx, data_size;
-	uint64_t *ts_end;
 
 	endidx = subbuf_index(offsets->end - 1, chan);
 	data_size = subbuf_offset(offsets->end - 1, chan) + 1;
 	subbuffer_set_data_size(config, &buf->backend, endidx, data_size,
 				handle);
-	ts_end = shmp_index(handle, buf->ts_end, endidx);
-	cc_cold = shmp_index(handle, buf->commit_cold, endidx);
-	if (!ts_end || !cc_cold)
+	if (!backend_pages)
 		return;
 	/*
 	 * This is the last space reservation in that sub-buffer before
 	 * it gets delivered. This provides exclusive access to write to
-	 * this sub-buffer's ts_end. There are also no concurrent
-	 * readers of that ts_end because delivery of that sub-buffer is
-	 * postponed until the commit counter is incremented for the
-	 * current space reservation.
+	 * this sub-buffer's timestamp_end. There are also no concurrent
+	 * readers of that timestamp_end because delivery of that
+	 * sub-buffer is postponed until the commit counter is
+	 * incremented for the current space reservation.
 	 */
-	*ts_end = ctx->priv->timestamp;
-	cc_cold->end_events_discarded =
+	backend_pages->timestamp_end = ctx->priv->timestamp;
+	backend_pages->end_events_discarded =
 		lib_ring_buffer_get_records_lost_full(config, ctx) +
 		lib_ring_buffer_get_records_lost_wrap(config, ctx) +
 		lib_ring_buffer_get_records_lost_big(config, ctx);
@@ -1950,6 +2066,7 @@ int lib_ring_buffer_try_switch_slow(enum switch_mode mode,
 				    struct lttng_ust_ring_buffer *buf,
 				    struct lttng_ust_ring_buffer_channel *chan,
 				    struct switch_offsets *offsets,
+				    unsigned long *old_pos,
 				    struct lttng_ust_ring_buffer_ctx *ctx,
 				    struct lttng_ust_shm_handle *handle)
 {
@@ -1958,6 +2075,8 @@ int lib_ring_buffer_try_switch_slow(enum switch_mode mode,
 
 	offsets->begin = v_read(config, &buf->offset);
 	offsets->old = offsets->begin;
+	if (old_pos)
+		*old_pos = offsets->old;
 	offsets->switch_old_start = 0;
 	off = subbuf_offset(offsets->begin, chan);
 
@@ -1985,6 +2104,7 @@ int lib_ring_buffer_try_switch_slow(enum switch_mode mode,
 	if (caa_unlikely(off == 0)) {
 		unsigned long sb_index, commit_count;
 		struct commit_counters_cold *cc_cold;
+		struct commit_counters_hot *cc_hot;
 
 		/*
 		 * We are performing a SWITCH_FLUSH. There may be concurrent
@@ -2001,6 +2121,9 @@ int lib_ring_buffer_try_switch_slow(enum switch_mode mode,
 
 		/* Test new buffer integrity */
 		sb_index = subbuf_index(offsets->begin, chan);
+		cc_hot = shmp_index(handle, buf->commit_hot, sb_index);
+		if (!cc_hot)
+			return -1;
 		cc_cold = shmp_index(handle, buf->commit_cold, sb_index);
 		if (!cc_cold)
 			return -1;
@@ -2009,7 +2132,7 @@ int lib_ring_buffer_try_switch_slow(enum switch_mode mode,
 		  (buf_trunc(offsets->begin, chan)
 		   >> chan->backend.num_subbuf_order)
 		  - (commit_count & chan->commit_count_mask);
-		if (caa_likely(reserve_commit_diff == 0)) {
+		if (caa_likely(reserve_commit_diff == 0) && caa_likely(v_read(config, &cc_hot->owner) == LTTNG_UST_ABI_OWNER_ID_UNSET)) {
 			/* Next subbuffer not being written to. */
 			if (caa_unlikely(config->mode != RING_BUFFER_OVERWRITE &&
 				subbuf_trunc(offsets->begin, chan)
@@ -2069,14 +2192,13 @@ int lib_ring_buffer_try_switch_slow(enum switch_mode mode,
  * from any CPU.
  */
 void lib_ring_buffer_switch_slow(struct lttng_ust_ring_buffer *buf, enum switch_mode mode,
-				 struct lttng_ust_shm_handle *handle)
+				 unsigned long *old_pos, struct lttng_ust_shm_handle *handle)
 {
 	struct lttng_ust_ring_buffer_channel *chan;
 	const struct lttng_ust_ring_buffer_config *config;
 	struct lttng_ust_ring_buffer_ctx_private ctx_priv;
 	struct lttng_ust_ring_buffer_ctx ctx;
 	struct switch_offsets offsets;
-	unsigned long oldidx;
 
 	ctx.priv = &ctx_priv;
 	chan = shmp(handle, buf->backend.chan);
@@ -2091,39 +2213,36 @@ void lib_ring_buffer_switch_slow(struct lttng_ust_ring_buffer *buf, enum switch_
 	 */
 	do {
 		if (lib_ring_buffer_try_switch_slow(mode, buf, chan, &offsets,
-						    &ctx, handle))
+						    old_pos, &ctx, handle))
 			return;	/* Switch not needed */
 	} while (v_cmpxchg(config, &buf->offset, offsets.old, offsets.end)
 		 != offsets.old);
 
+	/*
+	 * This might seem odd, but it will force to pass the early check in
+	 * `lib_ring_buffer_clear_owner_lazy_padding' while still doing the
+	 * second check.
+	 */
+	ctx_priv.reserve_then = offsets.old - 1;
+
+	/*
+	 * Observable side-effect: Reservation offset has been incremented.
+	 * Consumer's snapshots and other producers reservations are impacted.
+	 */
+	TESTPOINT("lib_ring_buffer_switch_slow_cmpxchg_succeed");
+
 	/*
 	 * Atomically update last_timestamp. This update races against concurrent
 	 * atomic updates, but the race will always cause supplementary full
-	 * timestamp records, never the opposite (missing a full timestamp
-	 * record when it would be needed).
+	 * timestamp record headers, never the opposite (missing a full
+	 * timestamp record header when it would be needed).
 	 */
-	save_last_timestamp(config, buf, ctx.priv->timestamp);
+	save_last_timestamp(config, chan, buf, ctx_priv.timestamp);
 
-	/*
-	 * Push the reader if necessary
-	 */
-	lib_ring_buffer_reserve_push_reader(buf, chan, offsets.old);
-
-	oldidx = subbuf_index(offsets.old, chan);
-	lib_ring_buffer_clear_noref(config, &buf->backend, oldidx, handle);
-
-	/*
-	 * May need to populate header start on SWITCH_FLUSH.
-	 */
-	if (offsets.switch_old_start) {
-		lib_ring_buffer_switch_old_start(buf, chan, &offsets, &ctx, handle);
-		offsets.old += config->cb.subbuffer_header_size();
-	}
-
-	/*
-	 * Switch old subbuffer.
-	 */
-	lib_ring_buffer_switch_old_end(buf, chan, &offsets, &ctx, handle);
+	lib_ring_buffer_try_clear_lazy_padding(config, chan, buf,
+					subbuf_index(offsets.old, chan),
+					handle,
+					&ctx);
 }
 
 static
@@ -2198,6 +2317,7 @@ retry:
 	if (caa_unlikely(offsets->switch_new_start)) {
 		unsigned long sb_index, commit_count;
 		struct commit_counters_cold *cc_cold;
+		struct commit_counters_hot *cc_hot;
 
 		/*
 		 * We are typically not filling the previous buffer completely.
@@ -2216,6 +2336,9 @@ retry:
 		 * are not seen reordered when updated by another CPU.
 		 */
 		cmm_smp_rmb();
+		cc_hot = shmp_index(handle, buf->commit_hot, sb_index);
+		if (!cc_hot)
+			return -1;
 		cc_cold = shmp_index(handle, buf->commit_cold, sb_index);
 		if (!cc_cold)
 			return -1;
@@ -2235,7 +2358,7 @@ retry:
 		  (buf_trunc(offsets->begin, chan)
 		   >> chan->backend.num_subbuf_order)
 		  - (commit_count & chan->commit_count_mask);
-		if (caa_likely(reserve_commit_diff == 0)) {
+		if (caa_likely(reserve_commit_diff == 0) && caa_likely(v_read(config, &cc_hot->owner) == LTTNG_UST_ABI_OWNER_ID_UNSET)) {
 			/* Next subbuffer not being written to. */
 			if (caa_unlikely(config->mode != RING_BUFFER_OVERWRITE &&
 				subbuf_trunc(offsets->begin, chan)
@@ -2335,13 +2458,12 @@ retry:
 	}
 	/*
 	 * Populate the records lost counters when the space reservation
-	 * may cause a sub-buffer switch.
+	 * may cause a sub-buffer switch. Due to lazy padding, we always
+	 * need to populate those.
 	 */
-	if (offsets->switch_new_end || offsets->switch_old_end) {
-		ctx_private->records_lost_full = v_read(config, &buf->records_lost_full);
-		ctx_private->records_lost_wrap = v_read(config, &buf->records_lost_wrap);
-		ctx_private->records_lost_big = v_read(config, &buf->records_lost_big);
-	}
+	ctx_private->records_lost_full = v_read(config, &buf->records_lost_full);
+	ctx_private->records_lost_wrap = v_read(config, &buf->records_lost_wrap);
+	ctx_private->records_lost_big = v_read(config, &buf->records_lost_big);
 	return 0;
 }
 
@@ -2374,14 +2496,44 @@ int lib_ring_buffer_reserve_slow(struct lttng_ust_ring_buffer_ctx *ctx,
 
 	offsets.size = 0;
 
-	do {
+	while (1) {
+
 		ret = lib_ring_buffer_try_reserve_slow(buf, chan, &offsets,
 						       ctx, client_ctx);
 		if (caa_unlikely(ret))
 			return ret;
-	} while (caa_unlikely(v_cmpxchg(config, &buf->offset, offsets.old,
-				    offsets.end)
-			  != offsets.old));
+
+		if (caa_unlikely(lib_ring_buffer_try_take_subbuf_ownership(
+						config, chan, buf,
+						subbuf_index(offsets.end - 1, chan),
+						handle))) {
+			lib_ring_buffer_switch_slow(buf, SWITCH_FLUSH, NULL, handle);
+			continue;
+		}
+
+		/*
+		 * Observable side-effect: Ownership of the sub-buffer is now taken.
+		 * Other producers reservations are impacted.
+		 */
+		TESTPOINT("lib_ring_buffer_reserve_slow_take_ownership_succeed");
+
+		if (caa_unlikely(v_cmpxchg(config, &buf->offset, offsets.old,
+				    offsets.end) != offsets.old)) {
+			lib_ring_buffer_clear_owner_lazy_padding(config, chan, buf,
+								subbuf_index(offsets.end - 1, chan),
+								handle,
+								ctx);
+			continue;
+		}
+
+		/*
+		 * Observable side-effect: Reservation offset has been incremented.
+		 * Consumer's snapshots and other producers reservations are impacted.
+		 */
+		TESTPOINT("lib_ring_buffer_reserve_slow_cmpxchg_succeed");
+
+		break;
+	}
 
 	/*
 	 * Atomically update last_timestamp. This update races against concurrent
@@ -2389,12 +2541,14 @@ int lib_ring_buffer_reserve_slow(struct lttng_ust_ring_buffer_ctx *ctx,
 	 * timestamp records, never the opposite (missing a full timestamp
 	 * record when it would be needed).
 	 */
-	save_last_timestamp(config, buf, ctx_private->timestamp);
+	save_last_timestamp(config, chan, buf, ctx_private->timestamp);
 
 	/*
 	 * Push the reader if necessary
 	 */
-	lib_ring_buffer_reserve_push_reader(buf, chan, offsets.end - 1);
+	lib_ring_buffer_reserve_push_reader(config, buf, chan, offsets.end - 1);
+
+	TESTPOINT("lib_ring_buffer_reserve_slow_after_push_reader");
 
 	/*
 	 * Clear noref flag for this subbuffer.
@@ -2407,12 +2561,24 @@ int lib_ring_buffer_reserve_slow(struct lttng_ust_ring_buffer_ctx *ctx,
 	 * Switch old subbuffer if needed.
 	 */
 	if (caa_unlikely(offsets.switch_old_end)) {
-		lib_ring_buffer_clear_noref(config, &buf->backend,
-					    subbuf_index(offsets.old - 1, chan),
-					    handle);
-		lib_ring_buffer_switch_old_end(buf, chan, &offsets, ctx, handle);
+		/*
+		 * This will force not to pass the first condition while still
+		 * checking the second condition in
+		 * lib_ring_buffer_lazy_padding_as_owner().
+		 */
+		ctx->priv->reserve_then = offsets.end - 1;
+
+		lib_ring_buffer_try_clear_lazy_padding(config, chan, buf,
+						subbuf_index(offsets.old - 1, chan),
+						handle, ctx);
 	}
 
+	/*
+	 * This is the actual reservation position to keep track of for this
+	 * commit.
+	 */
+	ctx->priv->reserve_then = offsets.end;
+
 	/*
 	 * Populate new subbuffer.
 	 */
@@ -2450,6 +2616,7 @@ void lib_ring_buffer_vmcore_check_deliver(const struct lttng_ust_ring_buffer_con
  * but it is disabled by default due to its performance overhead.
  */
 #ifdef LTTNG_RING_BUFFER_COUNT_EVENTS
+#error "The feature LTTNG_RING_BUFFER_COUNT_EVENTS is not compatible with stalled buffers fixup."
 static
 void deliver_count_events(const struct lttng_ust_ring_buffer_config *config,
 		struct lttng_ust_ring_buffer *buf,
@@ -2485,6 +2652,7 @@ void lib_ring_buffer_check_deliver_slow(const struct lttng_ust_ring_buffer_confi
 {
 	unsigned long old_commit_count = commit_count
 					 - chan->backend.subbuf_size;
+	struct lttng_ust_ring_buffer_backend_pages *backend_pages;
 	struct commit_counters_cold *cc_cold;
 
 	/*
@@ -2525,7 +2693,12 @@ void lib_ring_buffer_check_deliver_slow(const struct lttng_ust_ring_buffer_confi
 	if (caa_likely(v_cmpxchg(config, &cc_cold->cc_sb,
 				 old_commit_count, old_commit_count + 1)
 		   == old_commit_count)) {
-		uint64_t *ts_end;
+		/*
+		 * Observable side-effect: The sub-buffer's cold commit counter
+		 * has been incremented by one and the sub-buffer is now in
+		 * exlusived access by this producer.
+		 */
+		TESTPOINT("lib_ring_buffer_check_deliver_slow_cmpxchg_succeed");
 
 		/*
 		 * Start of exclusive subbuffer access. We are
@@ -2533,18 +2706,20 @@ void lib_ring_buffer_check_deliver_slow(const struct lttng_ust_ring_buffer_confi
 		 * and any other writer trying to access this subbuffer
 		 * in this state is required to drop records.
 		 *
-		 * We can read the ts_end for the current sub-buffer
+		 * We can read the timestamp_end for the current sub-buffer
 		 * which has been saved by the very last space
 		 * reservation for the current sub-buffer.
 		 *
-		 * Order increment of commit counter before reading ts_end.
+		 * Order increment of commit counter before reading
+		 * timestamp_end.
 		 */
 		cmm_smp_mb();
-		ts_end = shmp_index(handle, buf->ts_end, idx);
-		if (!ts_end)
+		backend_pages = lib_ring_buffer_offset_backend_pages(&buf->backend,
+					idx * chan->backend.subbuf_size, handle);
+		if (!backend_pages)
 			return;
 		deliver_count_events(config, buf, idx, handle);
-		config->cb.buffer_end(buf, *ts_end, idx,
+		config->cb.buffer_end(buf, backend_pages->timestamp_end, idx,
 				      lib_ring_buffer_get_data_size(config,
 								buf,
 								idx,
@@ -2555,25 +2730,37 @@ void lib_ring_buffer_check_deliver_slow(const struct lttng_ust_ring_buffer_confi
 		 * Increment the packet counter while we have exclusive
 		 * access.
 		 */
-		subbuffer_inc_packet_count(config, &buf->backend, idx, handle);
+		subbuffer_inc_packet_count(chan, buf, handle, commit_count, idx);
+
+		TESTPOINT("lib_ring_buffer_check_deliver_slow_before_set_noref");
 
 		/*
 		 * Set noref flag and offset for this subbuffer id.
 		 * Contains a memory barrier that ensures counter stores
 		 * are ordered before set noref and offset.
 		 */
-		lib_ring_buffer_set_noref_offset(config, &buf->backend, idx,
-						 buf_trunc_val(offset, chan), handle);
+		if (caa_unlikely(ctx->priv->is_fixup)) {
+			struct lttng_ust_ring_buffer_backend_subbuffer *wsb;
+			wsb = shmp_index(handle, buf->backend.buf_wsb, idx);
+			if (wsb && !subbuffer_id_is_noref(config, wsb->id)) {
+				lib_ring_buffer_set_noref_offset(config, &buf->backend, idx,
+							buf_trunc_val(offset - 1, chan), handle);
+			}
+
+		} else {
+			lib_ring_buffer_set_noref_offset(config, &buf->backend, idx,
+							buf_trunc_val(offset - 1, chan), handle);
+		}
 
 		/*
-		 * Order set_noref and record counter updates before the
-		 * end of subbuffer exclusive access. Orders with
-		 * respect to writers coming into the subbuffer after
-		 * wrap around, and also order wrt concurrent readers.
+		 * Store-release orders set_noref and record counter
+		 * updates before the end of subbuffer exclusive access.
+		 * Orders with respect to writers coming into the
+		 * subbuffer after wrap around, and also order wrt
+		 * concurrent readers.
 		 */
-		cmm_smp_mb();
 		/* End of exclusive subbuffer access */
-		v_set(config, &cc_cold->cc_sb, commit_count);
+		v_store(config, &cc_cold->cc_sb, commit_count, CMM_RELEASE);
 		/*
 		 * Order later updates to reserve count after
 		 * the commit cold cc_sb update.
@@ -2582,6 +2769,12 @@ void lib_ring_buffer_check_deliver_slow(const struct lttng_ust_ring_buffer_confi
 		lib_ring_buffer_vmcore_check_deliver(config, buf,
 					 commit_count, idx, handle);
 
+		/*
+		 * Observable side-effect: The consumer is notified by the
+		 * producer for a new sub-buffer to consume.
+		 */
+		TESTPOINT("lib_ring_buffer_check_deliver_slow_before_wakeup");
+
 		/*
 		 * RING_BUFFER_WAKEUP_BY_WRITER wakeup is not lock-free.
 		 */
@@ -2617,3 +2810,46 @@ void lib_ringbuffer_signal_init(void)
 		PERROR("pthread_sigmask");
 	}
 }
+
+void lib_ring_buffer_lazy_padding_as_owner_slow(const struct lttng_ust_ring_buffer_config *config,
+						struct lttng_ust_ring_buffer_channel *chan,
+						struct lttng_ust_ring_buffer *buf,
+						unsigned long subbuf_idx,
+						struct lttng_ust_shm_handle *handle,
+						const struct lttng_ust_ring_buffer_ctx *ctx,
+						unsigned long hot)
+{
+	struct switch_offsets offsets = { 0 };
+	unsigned long padding, oldidx;
+
+	assert(v_read(config, &buf->offset) - (long)subbuf_minimum_reserve(hot, subbuf_idx, chan) > 0);
+
+	padding = subbuf_trunc(hot, chan) + chan->backend.subbuf_size - hot;
+
+	/* The "reserve" value if hot was aligned to the next subbufer. */
+	offsets.end = (((subbuf_trunc(hot, chan) >> chan->backend.subbuf_size_order) << chan->backend.buf_size_order)
+		+ (subbuf_idx << chan->backend.subbuf_size_order) + chan->backend.subbuf_size);
+
+	offsets.old = offsets.begin = offsets.end - padding;
+	offsets.switch_old_start = subbuf_offset(offsets.begin, chan) == 0;
+
+	/*
+	 * Push the reader if necessary
+	 */
+	lib_ring_buffer_reserve_push_reader(config, buf, chan, offsets.end - 1);
+
+	oldidx = subbuf_index(offsets.old, chan);
+	assert(oldidx == subbuf_idx);
+
+	lib_ring_buffer_clear_noref(config, &buf->backend, oldidx, handle);
+
+	if (offsets.switch_old_start) {
+		lib_ring_buffer_switch_old_start(buf, chan, &offsets, ctx, handle);
+		offsets.old += config->cb.subbuffer_header_size();
+	}
+
+	/*
+	 * Switch old subbuffer.
+	 */
+	lib_ring_buffer_switch_old_end(buf, chan, &offsets, ctx, handle);
+}
