@@ -46,7 +46,9 @@ struct lttng_ust_ring_buffer_client_cb {
 
 	/* Slow path only, at subbuffer switch */
 	size_t (*subbuffer_header_size) (void);
-	void (*buffer_begin) (struct lttng_ust_ring_buffer *buf, uint64_t timestamp,
+	void (*buffer_begin) (struct lttng_ust_ring_buffer *buf,
+			      uint64_t timestamp,
+			      uint64_t discarded_events,
 			      unsigned int subbuf_idx,
 			      struct lttng_ust_shm_handle *handle);
 	void (*buffer_end) (struct lttng_ust_ring_buffer *buf, uint64_t timestamp,
@@ -86,6 +88,9 @@ struct lttng_ust_ring_buffer_client_cb {
 				size_t *offset, size_t *length);
 	void (*packet_size_field) (const struct lttng_ust_ring_buffer_config *config,
 				size_t *offset, size_t *length);
+
+	/* Get clock frequency for the channel. */
+	uint64_t (*ring_buffer_clock_freq) (struct lttng_ust_ring_buffer_channel *chan);
 };
 
 /*
