@@ -21,8 +21,8 @@
 #include <inttypes.h>
 
 #include <lttng/ust-ctl.h>
+#include <lttng/ust-fd.h>
 #include "common/ustcomm.h"
-#include "common/ust-fd.h"
 #include "common/macros.h"
 #include "common/dynamic-type.h"
 #include "common/logging.h"
@@ -30,8 +30,6 @@
 #include "common/events.h"
 #include "common/compat/pthread.h"
 
-#define USTCOMM_MAX_SEND_FDS	4
-
 static
 ssize_t count_fields_recursive(size_t nr_fields,
 		const struct lttng_ust_event_field * const *lttng_fields);
@@ -296,23 +294,20 @@ ssize_t ustcomm_recv_unix_sock(int sock, void *buf, size_t len)
 }
 
 /*
- * ustcomm_send_unix_sock
+ * ustcomm_sendv_unix_sock
  *
- * Send buf data of size len. Using sendmsg API.
+ * Send iov[iovcnt] over sock. Using sendmsg API.
  * Return the size of sent data.
  */
-ssize_t ustcomm_send_unix_sock(int sock, const void *buf, size_t len)
+ssize_t ustcomm_sendv_unix_sock(int sock, const struct iovec *iov, int iovcnt)
 {
 	struct msghdr msg;
-	struct iovec iov[1];
 	ssize_t ret;
 
 	memset(&msg, 0, sizeof(msg));
 
-	iov[0].iov_base = (void *) buf;
-	iov[0].iov_len = len;
-	msg.msg_iov = iov;
-	msg.msg_iovlen = 1;
+	msg.msg_iov = (struct iovec*)iov;
+	msg.msg_iovlen = iovcnt;
 
 	/*
 	 * Using the MSG_NOSIGNAL when sending data from sessiond to
@@ -338,6 +333,14 @@ ssize_t ustcomm_send_unix_sock(int sock, const void *buf, size_t len)
 	return ret;
 }
 
+ssize_t ustcomm_send_unix_sock(int sock, const void *buf, size_t len)
+{
+	struct iovec iov[1];
+	iov[0].iov_base = (void *) buf;
+	iov[0].iov_len = len;
+	return ustcomm_sendv_unix_sock(sock, iov, 1);
+}
+
 /*
  * Send a message accompanied by fd(s) over a unix socket.
  *
@@ -356,7 +359,7 @@ ssize_t ustcomm_send_fds_unix_sock(int sock, int *fds, size_t nb_fd)
 	memset(&msg, 0, sizeof(msg));
 	memset(tmp, 0, CMSG_SPACE(sizeof_fds) * sizeof(char));
 
-	if (nb_fd > USTCOMM_MAX_SEND_FDS)
+	if (nb_fd > LTTNG_UST_COMM_MAX_SEND_FDS)
 		return -EINVAL;
 
 	msg.msg_control = (caddr_t)tmp;
@@ -473,22 +476,141 @@ end:
 	return ret;
 }
 
-int ustcomm_send_app_msg(int sock, struct ustcomm_ust_msg *lum)
+static ssize_t
+prepare_ancillary_data(struct msghdr *msg, const int *fds, int fds_cnt, char *control_buf)
 {
-	ssize_t len;
+	size_t sizeof_fds = fds_cnt * sizeof(int);
+	struct cmsghdr *cmptr;
 
-	len = ustcomm_send_unix_sock(sock, lum, sizeof(*lum));
-	switch (len) {
-	case sizeof(*lum):
-		break;
-	default:
-		if (len < 0) {
-			return len;
-		} else {
-			ERR("incorrect message size: %zd\n", len);
+	if (!fds) {
+
+		if (fds_cnt != 0)
+			return -EINVAL;
+
+		msg->msg_control = NULL;
+		msg->msg_controllen = 0;
+
+		return 0;
+	}
+
+	if (fds_cnt <= 0 || fds_cnt > LTTNG_UST_COMM_MAX_SEND_FDS)
+		return -EINVAL;
+
+	msg->msg_control = control_buf;
+	msg->msg_controllen = CMSG_SPACE(sizeof_fds);
+
+	cmptr = CMSG_FIRSTHDR(msg);
+
+	cmptr->cmsg_level = SOL_SOCKET;
+	cmptr->cmsg_type = SCM_RIGHTS;
+	cmptr->cmsg_len = CMSG_LEN(sizeof_fds);
+	memcpy(CMSG_DATA(cmptr), fds, sizeof_fds);
+
+	return sizeof_fds;
+}
+
+static ssize_t
+prepare_payload_data(struct msghdr *msg, struct iovec *iov, int iovcnt)
+{
+	size_t total_size = 0;
+
+	if (!iov) {
+
+		if (iovcnt != 0)
+			return -EINVAL;
+
+		return 0;
+	}
+
+	if (iovcnt == 0)
+		return -EINVAL;
+
+	for (int i = 0; i < iovcnt; ++i) {
+		total_size += iov[i].iov_len;
+	}
+
+	msg->msg_iov = iov;
+	msg->msg_iovlen = iovcnt;
+
+	return total_size;
+}
+
+int ustcomm_send_app_msg(int sock, struct ustcomm_ust_msg_header *lum,
+			struct iovec *iov, int iovcnt,
+			const int *fds, size_t fds_cnt)
+{
+	struct msghdr msghdr;
+	ssize_t sent_size, payload_size, ancillary_size;
+	char control_buf[CMSG_SPACE(LTTNG_UST_COMM_MAX_SEND_FDS * sizeof(int))]
+		__attribute__((aligned(__alignof__(struct cmsghdr))));
+
+	/*
+	 * Memset to ensure all bytes of unions are zeroed.
+	 */
+	memset(&msghdr, 0, sizeof(msghdr));
+	memset(control_buf, 0, sizeof(control_buf));
+
+	payload_size = prepare_payload_data(&msghdr, iov, iovcnt);
+	if (payload_size < 0) {
+		ERR("error while preparing payload data: %zd\n", payload_size);
+		return (int)payload_size;
+	}
+	if (payload_size > LTTNG_UST_COMM_MAX_PAYLOAD_SIZE) {
+		ERR("payload data size is too big: %zd\n", payload_size);
+		return -E2BIG;
+	}
+
+	ancillary_size = prepare_ancillary_data(&msghdr, fds, fds_cnt, control_buf);
+
+	if (ancillary_size < 0) {
+		ERR("error while preparing ancillary data: %zd\n", ancillary_size);
+		return (int)ancillary_size;
+	}
+
+	if (ancillary_size > LTTNG_UST_COMM_MAX_PAYLOAD_SIZE) {
+		ERR("ancillary data size is too big: %zd\n", ancillary_size);
+		return -E2BIG;
+	}
+
+	if (ancillary_size > 0) {
+		if (payload_size == 0) {
+			ERR("ancillary size is not null but payload size is zero");
 			return -EINVAL;
 		}
 	}
+
+	lum->payload_size = (uint32_t)payload_size;
+	lum->ancillary_size = (uint32_t)ancillary_size;
+
+	sent_size = ustcomm_send_unix_sock(sock, lum, sizeof(*lum));
+
+	switch (sent_size) {
+	case sizeof(*lum):
+		break;
+	default:
+		if (sent_size < 0) {
+			return sent_size;
+		} else {
+			ERR("incorrect message size: %zd\n", sent_size);
+			return -EINVAL;
+		}
+	}
+
+	if (lum->payload_size == 0) {
+		return 0;
+	}
+
+	sent_size = sendmsg(sock, &msghdr, MSG_NOSIGNAL);
+
+	if (sent_size != lum->payload_size) {
+		if (sent_size < 0) {
+			return sent_size;
+		} else {
+			ERR("incorrect message size: %zd\n", sent_size);
+			return -EINVAL;
+		}
+	}
+
 	return 0;
 }
 
@@ -506,22 +628,22 @@ int ustcomm_recv_app_reply(int sock, struct ustcomm_ust_reply *lur,
 	{
 		int err = 0;
 
-		if (lur->handle != expected_handle) {
+		if (lur->header.handle != expected_handle) {
 			ERR("Unexpected result message handle: "
 				"expected: %u vs received: %u\n",
-				expected_handle, lur->handle);
+				expected_handle, lur->header.handle);
 			err = 1;
 		}
-		if (lur->cmd != expected_cmd) {
+		if (lur->header.cmd != expected_cmd) {
 			ERR("Unexpected result message command "
 				"expected: %u vs received: %u\n",
-				expected_cmd, lur->cmd);
+				expected_cmd, lur->header.cmd);
 			err = 1;
 		}
 		if (err) {
 			return -EINVAL;
 		} else {
-			return lur->ret_code;
+			return lur->header.ret_code;
 		}
 	}
 	default:
@@ -533,12 +655,14 @@ int ustcomm_recv_app_reply(int sock, struct ustcomm_ust_reply *lur,
 }
 
 int ustcomm_send_app_cmd(int sock,
-			struct ustcomm_ust_msg *lum,
+			struct ustcomm_ust_msg_header *lum,
 			struct ustcomm_ust_reply *lur)
 {
 	int ret;
 
-	ret = ustcomm_send_app_msg(sock, lum);
+	ret = ustcomm_send_app_msg(sock, lum,
+				NULL, 0,
+				NULL, 0);
 	if (ret)
 		return ret;
 	ret = ustcomm_recv_app_reply(sock, lur, lum->handle, lum->cmd);
@@ -547,236 +671,6 @@ int ustcomm_send_app_cmd(int sock,
 	return ret;
 }
 
-/*
- * chan_data is allocated internally if this function returns the
- * expected var_len.
- */
-ssize_t ustcomm_recv_channel_from_sessiond(int sock,
-		void **_chan_data, uint64_t var_len,
-		int *_wakeup_fd)
-{
-	void *chan_data;
-	ssize_t len, nr_fd;
-	int wakeup_fd, ret;
-
-	if (var_len > LTTNG_UST_ABI_CHANNEL_DATA_MAX_LEN) {
-		len = -EINVAL;
-		goto error_check;
-	}
-	/* Receive variable length data */
-	chan_data = zmalloc(var_len);
-	if (!chan_data) {
-		len = -ENOMEM;
-		goto error_alloc;
-	}
-	len = ustcomm_recv_unix_sock(sock, chan_data, var_len);
-	if (len != var_len) {
-		goto error_recv;
-	}
-	/* recv wakeup fd */
-	lttng_ust_lock_fd_tracker();
-	nr_fd = ustcomm_recv_fds_unix_sock(sock, &wakeup_fd, 1);
-	if (nr_fd <= 0) {
-		lttng_ust_unlock_fd_tracker();
-		if (nr_fd < 0) {
-			len = nr_fd;
-			goto error_recv;
-		} else {
-			len = -EIO;
-			goto error_recv;
-		}
-	}
-
-	ret = lttng_ust_add_fd_to_tracker(wakeup_fd);
-	if (ret < 0) {
-		ret = close(wakeup_fd);
-		if (ret) {
-			PERROR("close on wakeup_fd");
-		}
-		len = -EIO;
-		lttng_ust_unlock_fd_tracker();
-		goto error_recv;
-	}
-
-	*_wakeup_fd = ret;
-	lttng_ust_unlock_fd_tracker();
-
-	*_chan_data = chan_data;
-	return len;
-
-error_recv:
-	free(chan_data);
-error_alloc:
-error_check:
-	return len;
-}
-
-ssize_t ustcomm_recv_event_notifier_notif_fd_from_sessiond(int sock,
-		int *_event_notifier_notif_fd)
-{
-	ssize_t nr_fd;
-	int event_notifier_notif_fd, ret;
-
-	/* Receive event_notifier notification fd */
-	lttng_ust_lock_fd_tracker();
-	nr_fd = ustcomm_recv_fds_unix_sock(sock, &event_notifier_notif_fd, 1);
-	if (nr_fd <= 0) {
-		lttng_ust_unlock_fd_tracker();
-		if (nr_fd < 0) {
-			ret = nr_fd;
-			goto error;
-		} else {
-			ret = -EIO;
-			goto error;
-		}
-	}
-
-	ret = lttng_ust_add_fd_to_tracker(event_notifier_notif_fd);
-	if (ret < 0) {
-		ret = close(event_notifier_notif_fd);
-		if (ret) {
-			PERROR("close on event_notifier notif fd");
-		}
-		ret = -EIO;
-		lttng_ust_unlock_fd_tracker();
-		goto error;
-	}
-
-	*_event_notifier_notif_fd = ret;
-	lttng_ust_unlock_fd_tracker();
-
-	ret = nr_fd;
-
-error:
-	return ret;
-}
-
-int ustcomm_recv_stream_from_sessiond(int sock,
-		uint64_t *memory_map_size __attribute__((unused)),
-		int *shm_fd, int *wakeup_fd)
-{
-	ssize_t len;
-	int ret;
-	int fds[2];
-
-	/* recv shm fd and wakeup fd */
-	lttng_ust_lock_fd_tracker();
-	len = ustcomm_recv_fds_unix_sock(sock, fds, 2);
-	if (len <= 0) {
-		lttng_ust_unlock_fd_tracker();
-		if (len < 0) {
-			ret = len;
-			goto error;
-		} else {
-			ret = -EIO;
-			goto error;
-		}
-	}
-
-	ret = lttng_ust_add_fd_to_tracker(fds[0]);
-	if (ret < 0) {
-		ret = close(fds[0]);
-		if (ret) {
-			PERROR("close on received shm_fd");
-		}
-		ret = -EIO;
-		lttng_ust_unlock_fd_tracker();
-		goto error;
-	}
-	*shm_fd = ret;
-
-	ret = lttng_ust_add_fd_to_tracker(fds[1]);
-	if (ret < 0) {
-		ret = close(*shm_fd);
-		if (ret) {
-			PERROR("close on shm_fd");
-		}
-		*shm_fd = -1;
-		ret = close(fds[1]);
-		if (ret) {
-			PERROR("close on received wakeup_fd");
-		}
-		ret = -EIO;
-		lttng_ust_unlock_fd_tracker();
-		goto error;
-	}
-	*wakeup_fd = ret;
-	lttng_ust_unlock_fd_tracker();
-	return 0;
-
-error:
-	return ret;
-}
-
-ssize_t ustcomm_recv_var_len_cmd_from_sessiond(int sock,
-		void **_data, uint32_t var_len)
-{
-	void *data;
-	ssize_t len;
-
-	if (var_len > LTTNG_UST_ABI_CMD_MAX_LEN) {
-		len = -EINVAL;
-		goto error_check;
-	}
-	/* Receive variable length data */
-	data = zmalloc(var_len);
-	if (!data) {
-		len = -ENOMEM;
-		goto error_alloc;
-	}
-	len = ustcomm_recv_unix_sock(sock, data, var_len);
-	if (len != var_len) {
-		goto error_recv;
-	}
-	*_data = data;
-	return len;
-
-error_recv:
-	free(data);
-error_alloc:
-error_check:
-	return len;
-}
-
-int ustcomm_recv_counter_shm_from_sessiond(int sock,
-		int *shm_fd)
-{
-	ssize_t len;
-	int ret;
-	int fds[1];
-
-	/* recv shm fd fd */
-	lttng_ust_lock_fd_tracker();
-	len = ustcomm_recv_fds_unix_sock(sock, fds, 1);
-	if (len <= 0) {
-		lttng_ust_unlock_fd_tracker();
-		if (len < 0) {
-			ret = len;
-			goto error;
-		} else {
-			ret = -EIO;
-			goto error;
-		}
-	}
-
-	ret = lttng_ust_add_fd_to_tracker(fds[0]);
-	if (ret < 0) {
-		ret = close(fds[0]);
-		if (ret) {
-			PERROR("close on received shm_fd");
-		}
-		ret = -EIO;
-		lttng_ust_unlock_fd_tracker();
-		goto error;
-	}
-	*shm_fd = ret;
-	lttng_ust_unlock_fd_tracker();
-	return 0;
-
-error:
-	return ret;
-}
-
 /*
  * Returns 0 on success, negative error value on error.
  */
@@ -793,6 +687,7 @@ int ustcomm_send_reg_msg(int sock,
 	ssize_t len;
 	struct lttng_ust_ctl_reg_msg reg_msg;
 
+	memset(reg_msg.padding, 0, sizeof(reg_msg.padding));
 	reg_msg.magic = LTTNG_UST_ABI_COMM_MAGIC;
 	reg_msg.major = LTTNG_UST_ABI_MAJOR_VERSION;
 	reg_msg.minor = LTTNG_UST_ABI_MINOR_VERSION;
@@ -809,7 +704,6 @@ int ustcomm_send_reg_msg(int sock,
 	reg_msg.socket_type = type;
 	memset(reg_msg.name, 0, sizeof(reg_msg.name));
 	strncpy(reg_msg.name, procname, sizeof(reg_msg.name) - 1);
-	memset(reg_msg.padding, 0, sizeof(reg_msg.padding));
 
 	len = ustcomm_send_unix_sock(sock, &reg_msg, sizeof(reg_msg));
 	if (len > 0 && len != sizeof(reg_msg))
