@@ -18,45 +18,7 @@
 #include "lib/lttng-ust/events.h"
 #include "common/macros.h"
 #include "common/tracer.h"
-
-static int lttng_fls(int val)
-{
-	int r = 32;
-	unsigned int x = (unsigned int) val;
-
-	if (!x)
-		return 0;
-	if (!(x & 0xFFFF0000U)) {
-		x <<= 16;
-		r -= 16;
-	}
-	if (!(x & 0xFF000000U)) {
-		x <<= 8;
-		r -= 8;
-	}
-	if (!(x & 0xF0000000U)) {
-		x <<= 4;
-		r -= 4;
-	}
-	if (!(x & 0xC0000000U)) {
-		x <<= 2;
-		r -= 2;
-	}
-	if (!(x & 0x80000000U)) {
-		r -= 1;
-	}
-	return r;
-}
-
-static int get_count_order(unsigned int count)
-{
-	int order;
-
-	order = lttng_fls(count) - 1;
-	if (count & (count - 1))
-		order++;
-	return order;
-}
+#include "common/utils.h"
 
 static ssize_t bytecode_reserve_data(struct bytecode_runtime *runtime,
 		size_t align, size_t len)
@@ -74,7 +36,7 @@ static ssize_t bytecode_reserve_data(struct bytecode_runtime *runtime,
 		char *newptr;
 
 		new_alloc_len =
-			max_t(size_t, 1U << get_count_order(new_alloc_len), old_alloc_len << 1);
+			max_t(size_t, 1U << lttng_ust_get_count_order_ulong(new_alloc_len), old_alloc_len << 1);
 		newptr = realloc(runtime->data, new_alloc_len);
 		if (!newptr)
 			return -ENOMEM;
