@@ -35,8 +35,8 @@
 #include <lttng/ust-error.h>
 #include <lttng/ust-events.h>
 #include <lttng/ust-version.h>
+#include <lttng/ust-fd.h>
 
-#include "common/ust-fd.h"
 #include "common/logging.h"
 #include "common/align.h"
 
@@ -576,6 +576,11 @@ int lttng_abi_map_channel(int session_objd,
 		goto objd_error;
 	}
 
+	if (channel_set_owner_id(chan, ust_chan->owner_id)) {
+		ret = -EINVAL;
+		goto objd_error;
+	}
+
 	/* Initialize our lttng chan */
 	lttng_chan_buf->parent->enabled = 1;
 	lttng_chan_buf->parent->session = session;
@@ -594,7 +599,7 @@ int lttng_abi_map_channel(int session_objd,
 	lttng_chan_buf->priv->type = type;
 	/* Copy fields from lttng ust chan config. */
 	lttng_chan_buf->priv->id = lttng_chan_config->id;
-	memcpy(lttng_chan_buf->priv->uuid, lttng_chan_config->uuid, LTTNG_UST_UUID_LEN);
+	memcpy(lttng_chan_buf->priv->uuid, lttng_chan_config->uuid, LTTNG_UST_ABI_UUID_LEN);
 	channel_set_private(chan, lttng_chan_buf);
 
 	/*
@@ -849,7 +854,7 @@ static int lttng_ust_event_notifier_enabler_create(int event_notifier_group_obj,
 	struct lttng_event_notifier_enabler *event_notifier_enabler;
 	int event_notifier_objd, ret;
 
-	event_notifier_param->event.name[LTTNG_UST_ABI_SYM_NAME_LEN - 1] = '\0';
+	event_notifier_param->name[LTTNG_UST_ABI_SYM_NAME_LEN - 1] = '\0';
 	event_notifier_objd = objd_alloc(NULL, &lttng_event_notifier_enabler_ops, owner,
 		"event_notifier enabler");
 	if (event_notifier_objd < 0) {
@@ -1120,8 +1125,8 @@ long lttng_event_notifier_group_cmd(int objd, unsigned int cmd, unsigned long ar
 				abi_event_notifier, uargs->event_notifier.len);
 		if (ret)
 			return ret;
-		event_notifier.event.name[LTTNG_UST_ABI_SYM_NAME_LEN - 1] = '\0';
-		if (strutils_is_star_glob_pattern(event_notifier.event.name)) {
+		event_notifier.name[LTTNG_UST_ABI_SYM_NAME_LEN - 1] = '\0';
+		if (strutils_is_star_glob_pattern(event_notifier.name)) {
 			/*
 			 * If the event name is a star globbing pattern,
 			 * we create the special star globbing enabler.
