@@ -6,6 +6,7 @@
  */
 
 #define _LGPL_SOURCE
+#include <inttypes.h>
 #include <stddef.h>
 #include <stdint.h>
 #include <sys/types.h>
@@ -39,10 +40,10 @@
 #include <lttng/ust-tracer.h>
 #include <lttng/ust-common.h>
 #include <lttng/ust-cancelstate.h>
+#include <lttng/ust-fd.h>
 #include <urcu/tls-compat.h>
 #include "lib/lttng-ust/futex.h"
 #include "common/ustcomm.h"
-#include "common/ust-fd.h"
 #include "common/logging.h"
 #include "common/macros.h"
 #include "common/tracepoint.h"
@@ -59,6 +60,23 @@
 #include "common/counter-clients/clients.h"
 #include "common/ringbuffer-clients/clients.h"
 
+/*
+ * Becareful with side-effects and these macros.
+ */
+#define EXPECT_PAYLOAD_SIZE(payload_size, expected_size)		\
+	if ((payload_size) != (expected_size)) {			\
+		ERR("payload size `%zu` does not match expected size: %zu\n", \
+			(size_t)(payload_size), (size_t)(expected_size)); \
+	}								\
+	if ((payload_size) != (expected_size))
+
+#define EXPECT_ANCILLARY_SIZE(ancillary_size, expected_size)		\
+	if ((ancillary_size) != (expected_size)) {			\
+		ERR("ancillary size `%zu` does not match expected size: %zu\n", \
+			(size_t)(ancillary_size), (size_t)(expected_size)); \
+	}								\
+	if ((ancillary_size) != (expected_size))
+
 /*
  * Has lttng ust comm constructor been called ?
  */
@@ -277,7 +295,7 @@ struct sock_info {
 	int statedump_pending;
 	int initial_statedump_done;
 	/* Keep procname for statedump */
-	char procname[LTTNG_UST_CONTEXT_PROCNAME_LEN];
+	char procname[LTTNG_UST_ABI_PROCNAME_LEN];
 };
 
 /* Socket from app (connect) to session daemon (listen) for communication */
@@ -540,18 +558,27 @@ char* lttng_ust_sockinfo_get_procname(void *owner)
 	return info->procname;
 }
 
+/*
+ * Print the command name but also return true if the command is known.
+ */
 static
-void print_cmd(int cmd, int handle)
+bool print_cmd(int cmd, int handle, uint32_t payload_size, uint32_t ancillary_size)
 {
 	const char *cmd_name = "Unknown";
+	bool known = false;
 
-	if (cmd >= 0 && cmd < LTTNG_ARRAY_SIZE(cmd_name_mapping)
-			&& cmd_name_mapping[cmd]) {
-		cmd_name = cmd_name_mapping[cmd];
+	if (cmd >= 0 && cmd < LTTNG_ARRAY_SIZE(cmd_name_mapping)) {
+		if (cmd_name_mapping[cmd]) {
+			cmd_name = cmd_name_mapping[cmd];
+		}
+		known = true;
 	}
-	DBG("Message Received \"%s\" (%d), Handle \"%s\" (%d)",
+	DBG("Message Received \"%s\" (%d), Handle \"%s\" (%d), Payload Size (%" PRIu32 "), Ancillary Size (%" PRIu32 ")",
 		cmd_name, cmd,
-		lttng_ust_obj_get_name(handle), handle);
+		lttng_ust_obj_get_name(handle), handle,
+		payload_size, ancillary_size);
+
+	return known;
 }
 
 static
@@ -600,7 +627,7 @@ int setup_ust_apps(void)
 		goto end;
 	}
 
-	lttng_pthread_getname_np(ust_app.procname, LTTNG_UST_CONTEXT_PROCNAME_LEN);
+	lttng_pthread_getname_np(ust_app.procname, LTTNG_UST_ABI_PROCNAME_LEN);
 end:
 	if (ust_app_path)
 		free(ust_app_path);
@@ -629,7 +656,7 @@ int setup_global_apps(void)
 	}
 
 	global_apps.allowed = 1;
-	lttng_pthread_getname_np(global_apps.procname, LTTNG_UST_CONTEXT_PROCNAME_LEN);
+	lttng_pthread_getname_np(global_apps.procname, LTTNG_UST_ABI_PROCNAME_LEN);
 error:
 	return ret;
 }
@@ -683,7 +710,7 @@ int setup_local_apps(void)
 		goto end;
 	}
 
-	lttng_pthread_getname_np(local_apps.procname, LTTNG_UST_CONTEXT_PROCNAME_LEN);
+	lttng_pthread_getname_np(local_apps.procname, LTTNG_UST_ABI_PROCNAME_LEN);
 end:
 	return ret;
 }
@@ -907,72 +934,56 @@ const char *bytecode_type_str(uint32_t cmd)
 	}
 }
 
-enum handle_message_error {
-	MSG_OK = 0,
-	MSG_ERROR = 1,
-	MSG_SHUTDOWN = 2,
-};
-
-/*
- * Return:
- * < 0: error
- * 0: OK, handle command.
- * > 0: shutdown (no error).
- */
 static
-enum handle_message_error handle_error(struct sock_info *sock_info, ssize_t len,
-		ssize_t expected_len, const char *str, int *error_code)
-{
-	if (!len) {
-		/* orderly shutdown */
-		*error_code = 0;
-		return MSG_SHUTDOWN;
-	}
-	if (len == expected_len) {
-		DBG("%s data received", str);
-		*error_code = 0;
-		return MSG_OK;
-	}
-	if (len < 0) {
-		DBG("Receive failed from lttng-sessiond with errno %d", (int) -len);
-		if (len == -ECONNRESET) {
-			ERR("%s remote end closed connection", sock_info->name);
-		}
-		*error_code = len;
-		return MSG_ERROR;
-	}
-	DBG("incorrect %s data message size: %zd", str, len);
-	*error_code = -EINVAL;
-	return MSG_ERROR;
-}
-
-static
-int handle_bytecode_recv(struct sock_info *sock_info,
-		int sock, struct ustcomm_ust_msg *lum)
+int handle_bytecode(struct sock_info *sock_info,
+		struct ustcomm_ust_msg_header *lum,
+		const void *payload, uint32_t payload_size)
 {
 	struct lttng_ust_bytecode_node *bytecode = NULL;
 	enum lttng_ust_bytecode_type type;
 	const struct lttng_ust_abi_objd_ops *ops;
 	uint32_t data_size, data_size_max, reloc_offset;
+	const char *data;
 	uint64_t seqnum;
-	ssize_t len;
 	int ret = 0;
 
 	switch (lum->cmd) {
-	case LTTNG_UST_ABI_FILTER:
+	case LTTNG_UST_ABI_FILTER: {
+		const struct lttng_ust_abi_filter_bytecode *filter;
+
+		filter = payload;
+
+		EXPECT_PAYLOAD_SIZE(payload_size, sizeof(*filter) + filter->len) {
+			return -EINVAL;
+		}
+
 		type = LTTNG_UST_BYTECODE_TYPE_FILTER;
-		data_size = lum->u.filter.data_size;
 		data_size_max = LTTNG_UST_ABI_FILTER_BYTECODE_MAX_LEN;
-		reloc_offset = lum->u.filter.reloc_offset;
-		seqnum = lum->u.filter.seqnum;
+
+		data_size = filter->len;
+		reloc_offset = filter->reloc_offset;
+		seqnum = filter->seqnum;
+		data = filter->data;
 		break;
-	case LTTNG_UST_ABI_CAPTURE:
+	}
+	case LTTNG_UST_ABI_CAPTURE: {
+		const struct lttng_ust_abi_capture_bytecode *capture;
+
+		capture = payload;
+
+		EXPECT_PAYLOAD_SIZE(payload_size, sizeof(*capture) + capture->len) {
+			return -EINVAL;
+		}
+
 		type = LTTNG_UST_BYTECODE_TYPE_CAPTURE;
-		data_size = lum->u.capture.data_size;
 		data_size_max = LTTNG_UST_ABI_CAPTURE_BYTECODE_MAX_LEN;
-		reloc_offset = lum->u.capture.reloc_offset;
-		seqnum = lum->u.capture.seqnum;
+
+		data_size = capture->len;
+		reloc_offset = capture->reloc_offset;
+		seqnum = capture->seqnum;
+		data = capture->data;
 		break;
+	}
 	default:
 		abort();
 	}
@@ -1003,20 +1014,18 @@ int handle_bytecode_recv(struct sock_info *sock_info,
 	bytecode->bc.seqnum = seqnum;
 	bytecode->type = type;
 
-	len = ustcomm_recv_unix_sock(sock, bytecode->bc.data, bytecode->bc.len);
-	switch (handle_error(sock_info, len, bytecode->bc.len, bytecode_type_str(lum->cmd), &ret)) {
-	case MSG_OK:
-		break;
-	case MSG_ERROR:		/* Fallthrough */
-	case MSG_SHUTDOWN:
-		goto end;
-	}
+	memcpy(bytecode->bc.data, data, data_size);
+
 	ops = lttng_ust_abi_objd_ops(lum->handle);
 	if (!ops) {
 		ret = -ENOENT;
 		goto end;
 	}
 
+	/*
+	 * On success, the bytecode ownership is transfered and `bytecode' is
+	 * NULL.
+	 */
 	if (ops->cmd)
 		ret = ops->cmd(lum->handle, lum->cmd,
 			(unsigned long) &bytecode,
@@ -1032,11 +1041,11 @@ end:
 static
 void prepare_cmd_reply(struct ustcomm_ust_reply *lur, uint32_t handle, uint32_t cmd, int ret)
 {
-	lur->handle = handle;
-	lur->cmd = cmd;
-	lur->ret_val = ret;
+	lur->header.handle = handle;
+	lur->header.cmd = cmd;
+	lur->header.ret_val = ret;
 	if (ret >= 0) {
-		lur->ret_code = LTTNG_UST_OK;
+		lur->header.ret_code = LTTNG_UST_OK;
 	} else {
 		/*
 		 * Use -LTTNG_UST_ERR as wildcard for UST internal
@@ -1048,33 +1057,94 @@ void prepare_cmd_reply(struct ustcomm_ust_reply *lur, uint32_t handle, uint32_t
 			/* Translate code to UST error. */
 			switch (ret) {
 			case -EEXIST:
-				lur->ret_code = -LTTNG_UST_ERR_EXIST;
+				lur->header.ret_code = -LTTNG_UST_ERR_EXIST;
 				break;
 			case -EINVAL:
-				lur->ret_code = -LTTNG_UST_ERR_INVAL;
+				lur->header.ret_code = -LTTNG_UST_ERR_INVAL;
 				break;
 			case -ENOENT:
-				lur->ret_code = -LTTNG_UST_ERR_NOENT;
+				lur->header.ret_code = -LTTNG_UST_ERR_NOENT;
 				break;
 			case -EPERM:
-				lur->ret_code = -LTTNG_UST_ERR_PERM;
+				lur->header.ret_code = -LTTNG_UST_ERR_PERM;
 				break;
 			case -ENOSYS:
-				lur->ret_code = -LTTNG_UST_ERR_NOSYS;
+				lur->header.ret_code = -LTTNG_UST_ERR_NOSYS;
 				break;
 			default:
-				lur->ret_code = -LTTNG_UST_ERR;
+				lur->header.ret_code = -LTTNG_UST_ERR;
 				break;
 			}
 		} else {
-			lur->ret_code = ret;
+			lur->header.ret_code = ret;
 		}
 	}
 }
 
+static
+int copy_payload(const char *payload, uint32_t payload_size,
+		void **pcopy)
+{
+	void *copy;
+
+	if (payload_size > LTTNG_UST_ABI_CMD_MAX_LEN) {
+		return -EINVAL;
+	}
+
+	copy = zmalloc(payload_size);
+
+	if (!copy)
+		return -ENOMEM;
+
+	memcpy(copy, payload, payload_size);
+
+	*pcopy = copy;
+
+	return 0;
+}
+
+static
+int track_fd(int fd)
+{
+	int ret;
+	lttng_ust_lock_fd_tracker();
+	ret = lttng_ust_add_fd_to_tracker(fd);
+	if (ret < 0) {
+		ret = close(fd);
+		if (ret) {
+			PERROR("close on fd");
+		}
+		ret = -EIO;
+		lttng_ust_unlock_fd_tracker();
+	}
+	lttng_ust_unlock_fd_tracker();
+
+	return ret;
+}
+
+static
+void untrack_fd(int *fd)
+{
+	int close_ret;
+
+	if (*fd < 0)
+		return;
+
+	lttng_ust_lock_fd_tracker();
+	close_ret = lttng_ust_add_fd_to_tracker(*fd);
+	lttng_ust_unlock_fd_tracker();
+
+	*fd = -1;
+
+	if (close_ret)
+		PERROR("close on fd");
+}
+
 static
 int handle_message(struct sock_info *sock_info,
-		int sock, struct ustcomm_ust_msg *lum)
+		int sock, struct ustcomm_ust_msg_header *lum,
+		const void *restrict payload, size_t payload_size,
+		const void *restrict ancillary_data, size_t ancillary_size)
 {
 	int ret = 0;
 	const struct lttng_ust_abi_objd_ops *ops;
@@ -1083,6 +1153,7 @@ int handle_message(struct sock_info *sock_info,
 	char ctxstr[LTTNG_UST_ABI_SYM_NAME_LEN];	/* App context string. */
 	ssize_t len;
 	void *var_len_cmd_data = NULL;
+	const int *ancillary_fds = (const int *)ancillary_data;
 
 	if (ust_lock()) {
 		ret = -LTTNG_UST_ERR_EXITING;
@@ -1095,55 +1166,6 @@ int handle_message(struct sock_info *sock_info,
 		goto error;
 	}
 
-	switch (lum->cmd) {
-	case LTTNG_UST_ABI_FILTER:
-	case LTTNG_UST_ABI_EXCLUSION:
-	case LTTNG_UST_ABI_CHANNEL:
-	case LTTNG_UST_ABI_STREAM:
-	case LTTNG_UST_ABI_CONTEXT:
-		/*
-		 * Those commands send additional payload after struct
-		 * ustcomm_ust_msg, which makes it pretty much impossible to
-		 * deal with "unknown command" errors without leaving the
-		 * communication pipe in a out-of-sync state. This is part of
-		 * the ABI between liblttng-ust-ctl and liblttng-ust, and
-		 * should be fixed on the next breaking
-		 * LTTNG_UST_ABI_MAJOR_VERSION protocol bump by indicating the
-		 * total command message length as part of a message header so
-		 * that the protocol can recover from invalid command errors.
-		 */
-		break;
-
-	case LTTNG_UST_ABI_CAPTURE:
-	case LTTNG_UST_ABI_COUNTER:
-	case LTTNG_UST_ABI_COUNTER_CHANNEL:
-	case LTTNG_UST_ABI_COUNTER_CPU:
-#ifdef CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER
-	case LTTNG_UST_ABI_COUNTER_EVENT:
-#endif	/* CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER */
-	case LTTNG_UST_ABI_EVENT_NOTIFIER_CREATE:
-	case LTTNG_UST_ABI_EVENT_NOTIFIER_GROUP_CREATE:
-		/*
-		 * Those commands expect a reply to the struct ustcomm_ust_msg
-		 * before sending additional payload.
-		 */
-		prepare_cmd_reply(&lur, lum->handle, lum->cmd, 0);
-
-		ret = send_reply(sock, &lur);
-		if (ret < 0) {
-			DBG("error sending reply");
-			goto error;
-		}
-		break;
-
-	default:
-		/*
-		 * Other commands either don't send additional payload, or are
-		 * unknown.
-		 */
-		break;
-	}
-
 	switch (lum->cmd) {
 	case LTTNG_UST_ABI_REGISTER_DONE:
 		if (lum->handle == LTTNG_UST_ABI_ROOT_HANDLE)
@@ -1159,39 +1181,43 @@ int handle_message(struct sock_info *sock_info,
 		break;
 	case LTTNG_UST_ABI_CAPTURE:
 	case LTTNG_UST_ABI_FILTER:
-		ret = handle_bytecode_recv(sock_info, sock, lum);
+		ret = handle_bytecode(sock_info, lum, payload, payload_size);
 		if (ret)
 			goto error;
 		break;
 	case LTTNG_UST_ABI_EXCLUSION:
 	{
 		/* Receive exclusion names */
+		const struct lttng_ust_abi_event_exclusion *exclusion;
 		struct lttng_ust_excluder_node *node;
 		unsigned int count;
+		size_t exclusion_size;
+
+		exclusion = payload;
+
+		EXPECT_PAYLOAD_SIZE(payload_size, sizeof(*exclusion) +
+				exclusion->count * LTTNG_UST_ABI_SYM_NAME_LEN) {
+			ret = -EINVAL;
+			goto error;
+		}
+
+		count = exclusion->count;
+		exclusion_size = count * LTTNG_UST_ABI_SYM_NAME_LEN;
 
-		count = lum->u.exclusion.count;
 		if (count == 0) {
 			/* There are no names to read */
 			ret = 0;
 			goto error;
 		}
-		node = zmalloc(sizeof(*node) +
-				count * LTTNG_UST_ABI_SYM_NAME_LEN);
+		node = zmalloc(sizeof(*node) + exclusion_size);
 		if (!node) {
 			ret = -ENOMEM;
 			goto error;
 		}
 		node->excluder.count = count;
-		len = ustcomm_recv_unix_sock(sock, node->excluder.names,
-				count * LTTNG_UST_ABI_SYM_NAME_LEN);
-		switch (handle_error(sock_info, len, count * LTTNG_UST_ABI_SYM_NAME_LEN, "exclusion", &ret)) {
-		case MSG_OK:
-			break;
-		case MSG_ERROR:		/* Fallthrough */
-		case MSG_SHUTDOWN:
-			free(node);
-			goto error;
-		}
+
+		memcpy(node->excluder.names, exclusion->names, exclusion_size);
+
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
 					(unsigned long) &node,
@@ -1203,160 +1229,205 @@ int handle_message(struct sock_info *sock_info,
 	}
 	case LTTNG_UST_ABI_EVENT_NOTIFIER_GROUP_CREATE:
 	{
-		int event_notifier_notif_fd, close_ret;
+		int event_notifier_notif_fd;
 
-		len = ustcomm_recv_event_notifier_notif_fd_from_sessiond(sock,
-			&event_notifier_notif_fd);
-		switch (handle_error(sock_info, len, 1, "event notifier group", &ret)) {
-		case MSG_OK:
-			break;
-		case MSG_ERROR:		/* Fallthrough */
-		case MSG_SHUTDOWN:
+		EXPECT_PAYLOAD_SIZE(payload_size, 1) {
+			ret = -EINVAL;
 			goto error;
 		}
+
+		EXPECT_ANCILLARY_SIZE(ancillary_size, sizeof(int)) {
+			ret = -EINVAL;
+			goto error;
+		}
+
+		event_notifier_notif_fd = track_fd(ancillary_fds[0]);
+
+		if (event_notifier_notif_fd < 0) {
+			ret = event_notifier_notif_fd;
+			goto error;
+		}
+
 		args.event_notifier_handle.event_notifier_notif_fd = event_notifier_notif_fd;
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
-					(unsigned long) &lum->u,
+					(unsigned long) payload,
 					&args, sock_info);
 		else
 			ret = -ENOSYS;
-		if (args.event_notifier_handle.event_notifier_notif_fd >= 0) {
-			lttng_ust_lock_fd_tracker();
-			close_ret = close(args.event_notifier_handle.event_notifier_notif_fd);
-			lttng_ust_unlock_fd_tracker();
-			if (close_ret)
-				PERROR("close");
-		}
+		untrack_fd(&args.event_notifier_handle.event_notifier_notif_fd);
 		break;
 	}
 	case LTTNG_UST_ABI_CHANNEL:
 	{
-		void *chan_data;
 		int wakeup_fd;
+		void *chan_data;
+		const struct lttng_ust_abi_channel *channel = payload;
 
-		len = ustcomm_recv_channel_from_sessiond(sock,
-				&chan_data, lum->u.channel.len,
-				&wakeup_fd);
-		switch (handle_error(sock_info, len, lum->u.channel.len, "channel", &ret)) {
-		case MSG_OK:
-			break;
-		case MSG_ERROR:		/* Fallthrough */
-		case MSG_SHUTDOWN:
+		EXPECT_PAYLOAD_SIZE(payload_size, sizeof(*channel) + channel->len) {
+			ret = -EINVAL;
 			goto error;
 		}
+
+		EXPECT_ANCILLARY_SIZE(ancillary_size, sizeof(int)) {
+			ret = -EINVAL;
+			goto error;
+		}
+
+		wakeup_fd = track_fd(ancillary_fds[0]);
+
+		if (wakeup_fd < 0) {
+			ret = wakeup_fd;
+			goto error;
+		}
+
+		chan_data = zmalloc(payload_size);
+
+		if (!chan_data) {
+			untrack_fd(&wakeup_fd);
+			ret = -ENOMEM;
+			goto error;
+		}
+
+		memcpy(chan_data, channel->data, channel->len);
+
 		args.channel.chan_data = chan_data;
 		args.channel.wakeup_fd = wakeup_fd;
+
+		/* Chan data is copied. */
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
-					(unsigned long) &lum->u,
+					(unsigned long) payload,
 					&args, sock_info);
 		else
 			ret = -ENOSYS;
-		if (args.channel.wakeup_fd >= 0) {
-			int close_ret;
-
-			lttng_ust_lock_fd_tracker();
-			close_ret = close(args.channel.wakeup_fd);
-			lttng_ust_unlock_fd_tracker();
-			args.channel.wakeup_fd = -1;
-			if (close_ret)
-				PERROR("close");
-		}
-		free(args.channel.chan_data);
+		untrack_fd(&args.channel.wakeup_fd);
 		break;
 	}
 	case LTTNG_UST_ABI_STREAM:
 	{
-		int close_ret;
+		EXPECT_PAYLOAD_SIZE(payload_size, sizeof(struct lttng_ust_abi_stream)) {
+			ret = -EBADMSG;
+			goto error;
+		}
 
-		/* Receive shm_fd, wakeup_fd */
-		ret = ustcomm_recv_stream_from_sessiond(sock,
-			NULL,
-			&args.stream.shm_fd,
-			&args.stream.wakeup_fd);
-		if (ret) {
+
+		EXPECT_PAYLOAD_SIZE(ancillary_size, 2 * sizeof(int)) {
+			ret = -EBADMSG;
+			goto error;
+		}
+
+		args.stream.shm_fd = track_fd(ancillary_fds[0]);
+
+		if (args.stream.shm_fd < 0) {
+			ret = args.stream.shm_fd;
+			goto error;
+		}
+
+		args.stream.wakeup_fd = track_fd(ancillary_fds[1]);
+
+		if (args.stream.wakeup_fd < 0) {
+			untrack_fd(&args.stream.shm_fd);
+			ret = args.stream.wakeup_fd;
 			goto error;
 		}
 
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
-					(unsigned long) &lum->u,
+					(unsigned long) payload,
 					&args, sock_info);
 		else
 			ret = -ENOSYS;
-		if (args.stream.shm_fd >= 0) {
-			lttng_ust_lock_fd_tracker();
-			close_ret = close(args.stream.shm_fd);
-			lttng_ust_unlock_fd_tracker();
-			args.stream.shm_fd = -1;
-			if (close_ret)
-				PERROR("close");
-		}
-		if (args.stream.wakeup_fd >= 0) {
-			lttng_ust_lock_fd_tracker();
-			close_ret = close(args.stream.wakeup_fd);
-			lttng_ust_unlock_fd_tracker();
-			args.stream.wakeup_fd = -1;
-			if (close_ret)
-				PERROR("close");
-		}
+		untrack_fd(&args.stream.shm_fd);
+		untrack_fd(&args.stream.wakeup_fd);
 		break;
 	}
-	case LTTNG_UST_ABI_CONTEXT:
-		switch (lum->u.context.ctx) {
+	case LTTNG_UST_ABI_CONTEXT: {
+		const struct lttng_ust_abi_context *context = payload;
+
+		switch (context->header.ctx) {
 		case LTTNG_UST_ABI_CONTEXT_APP_CONTEXT:
 		{
-			char *p;
-			size_t ctxlen, recvlen;
+			const char *provider_name;
+			const char *context_name;
+			char *app_ctx = ctxstr;
+			size_t ctxlen;
 
-			ctxlen = strlen("$app.") + lum->u.context.u.app_ctx.provider_name_len - 1
-					+ strlen(":") + lum->u.context.u.app_ctx.ctx_name_len;
+			EXPECT_PAYLOAD_SIZE(payload_size, sizeof(*context) +
+					context->type.app_ctx.provider_name_len +
+					context->type.app_ctx.ctx_name_len) {
+				ret = -EBADMSG;
+				goto error;
+			}
+
+			EXPECT_ANCILLARY_SIZE(ancillary_size, 0) {
+				ret = -EBADMSG;
+				goto error;
+			}
+
+			if (context->type.app_ctx.provider_name_len + context->type.app_ctx.ctx_name_len >= LTTNG_UST_ABI_SYM_NAME_LEN) {
+				ERR("provider name and context name sizes `%" PRIu32 "`, `%" PRIu32 "` exceed maximum symbol name length: %u\n",
+					context->type.app_ctx.provider_name_len,
+					context->type.app_ctx.ctx_name_len,
+					LTTNG_UST_ABI_SYM_NAME_LEN);
+				ret = -EINVAL;
+				goto error;
+			}
+
+			/* NOTE: These are not null terminated strings. */
+			provider_name = payload + sizeof(*context);
+			context_name = provider_name + context->type.app_ctx.provider_name_len;
+
+			ctxlen = strlen("$app.") + context->type.app_ctx.provider_name_len
+				+ strlen(":") + context->type.app_ctx.ctx_name_len;
 			if (ctxlen >= LTTNG_UST_ABI_SYM_NAME_LEN) {
 				ERR("Application context string length size is too large: %zu bytes",
 					ctxlen);
 				ret = -EINVAL;
 				goto error;
 			}
-			strcpy(ctxstr, "$app.");
-			p = &ctxstr[strlen("$app.")];
-			recvlen = ctxlen - strlen("$app.");
-			len = ustcomm_recv_unix_sock(sock, p, recvlen);
-			switch (handle_error(sock_info, len, recvlen, "app context", &ret)) {
-			case MSG_OK:
-				break;
-			case MSG_ERROR:		/* Fallthrough */
-			case MSG_SHUTDOWN:
-				goto error;
-			}
-			/* Put : between provider and ctxname. */
-			p[lum->u.context.u.app_ctx.provider_name_len - 1] = ':';
+
+			strcpy(app_ctx, "$app.");
+			app_ctx += strlen("$app.");
+
+			memcpy(app_ctx, provider_name, context->type.app_ctx.provider_name_len);
+			app_ctx += context->type.app_ctx.provider_name_len;
+
+			app_ctx[0] = ':';
+			app_ctx += 1;
+
+			memcpy(app_ctx, context_name, context->type.app_ctx.ctx_name_len);
+			app_ctx += context->type.app_ctx.ctx_name_len;
+
+			app_ctx[0] = '\0';
+
 			args.app_context.ctxname = ctxstr;
 			break;
 		}
 		default:
+		{
+
 			break;
 		}
+		}
+
 		if (ops->cmd) {
 			ret = ops->cmd(lum->handle, lum->cmd,
-					(unsigned long) &lum->u,
+					(unsigned long) payload,
 					&args, sock_info);
 		} else {
 			ret = -ENOSYS;
 		}
 		break;
+	}
 	case LTTNG_UST_ABI_COUNTER:
 	{
-		len = ustcomm_recv_var_len_cmd_from_sessiond(sock,
-				&var_len_cmd_data, lum->u.var_len_cmd.cmd_len);
-		switch (handle_error(sock_info, len, lum->u.var_len_cmd.cmd_len, "counter", &ret)) {
-		case MSG_OK:
-			break;
-		case MSG_ERROR:		/* Fallthrough */
-		case MSG_SHUTDOWN:
+		ret = copy_payload(payload, payload_size, &var_len_cmd_data);
+
+		if (ret)
 			goto error;
-		}
-		args.counter.len = lum->u.var_len_cmd.cmd_len;
+
+		args.counter.len = payload_size;
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
 					(unsigned long) var_len_cmd_data,
@@ -1367,87 +1438,90 @@ int handle_message(struct sock_info *sock_info,
 	}
 	case LTTNG_UST_ABI_COUNTER_CHANNEL:
 	{
-		len = ustcomm_recv_var_len_cmd_from_sessiond(sock,
-				&var_len_cmd_data, lum->u.var_len_cmd.cmd_len);
-		switch (handle_error(sock_info, len, lum->u.var_len_cmd.cmd_len, "counter channel", &ret)) {
-		case MSG_OK:
-			break;
-		case MSG_ERROR:		/* Fallthrough */
-		case MSG_SHUTDOWN:
+		EXPECT_PAYLOAD_SIZE(payload_size, sizeof(struct lttng_ust_abi_counter_channel)) {
+			ret = -EBADMSG;
 			goto error;
 		}
-		/* Receive shm_fd */
-		ret = ustcomm_recv_counter_shm_from_sessiond(sock, &args.counter_shm.shm_fd);
+
+		EXPECT_ANCILLARY_SIZE(ancillary_size, sizeof(int)) {
+			ret = -EBADMSG;
+			goto error;
+		}
+
+		args.counter_shm.shm_fd = track_fd(ancillary_fds[0]);
+
+		if (args.counter_shm.shm_fd < 0) {
+			ret = args.counter_shm.shm_fd;
+			goto error;
+		}
+
+		ret = copy_payload(payload, payload_size,
+				&var_len_cmd_data);
+
 		if (ret) {
+			untrack_fd(&args.counter_shm.shm_fd);
 			goto error;
 		}
-		args.counter_shm.len = lum->u.var_len_cmd.cmd_len;
+
+		args.counter_shm.len = payload_size;
+
+		/* Take ownership of data. */
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
 					(unsigned long) var_len_cmd_data,
 					&args, sock_info);
 		else
 			ret = -ENOSYS;
-		if (args.counter_shm.shm_fd >= 0) {
-			int close_ret;
-
-			lttng_ust_lock_fd_tracker();
-			close_ret = close(args.counter_shm.shm_fd);
-			lttng_ust_unlock_fd_tracker();
-			args.counter_shm.shm_fd = -1;
-			if (close_ret)
-				PERROR("close");
-		}
+		untrack_fd(&args.counter_shm.shm_fd);
 		break;
 	}
 	case LTTNG_UST_ABI_COUNTER_CPU:
 	{
-		len = ustcomm_recv_var_len_cmd_from_sessiond(sock,
-				&var_len_cmd_data, lum->u.var_len_cmd.cmd_len);
-		switch (handle_error(sock_info, len, lum->u.var_len_cmd.cmd_len, "counter cpu", &ret)) {
-		case MSG_OK:
-			break;
-		case MSG_ERROR:		/* Fallthrough */
-		case MSG_SHUTDOWN:
+		EXPECT_PAYLOAD_SIZE(payload_size, sizeof(struct lttng_ust_abi_counter_cpu)) {
+			ret = -EBADMSG;
 			goto error;
 		}
-		/* Receive shm_fd */
-		ret = ustcomm_recv_counter_shm_from_sessiond(sock, &args.counter_shm.shm_fd);
+
+		EXPECT_ANCILLARY_SIZE(ancillary_size, sizeof(int)) {
+			ret = -EBADMSG;
+			goto error;
+		}
+
+		args.counter_shm.shm_fd = track_fd(ancillary_fds[0]);
+
+		if (args.counter_shm.shm_fd < 0) {
+			ret = args.counter_shm.shm_fd;
+			goto error;
+		}
+
+		ret = copy_payload(payload, payload_size,
+				&var_len_cmd_data);
+
 		if (ret) {
+			untrack_fd(&args.counter_shm.shm_fd);
 			goto error;
 		}
-		args.counter_shm.len = lum->u.var_len_cmd.cmd_len;
+
+		args.counter_shm.len = payload_size;
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
 					(unsigned long) var_len_cmd_data,
 					&args, sock_info);
 		else
 			ret = -ENOSYS;
-		if (args.counter_shm.shm_fd >= 0) {
-			int close_ret;
-
-			lttng_ust_lock_fd_tracker();
-			close_ret = close(args.counter_shm.shm_fd);
-			lttng_ust_unlock_fd_tracker();
-			args.counter_shm.shm_fd = -1;
-			if (close_ret)
-				PERROR("close");
-		}
+		untrack_fd(&args.counter_shm.shm_fd);
 		break;
 	}
 #ifdef CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER
 	case LTTNG_UST_ABI_COUNTER_EVENT:
 	{
-		len = ustcomm_recv_var_len_cmd_from_sessiond(sock,
-				&var_len_cmd_data, lum->u.var_len_cmd.cmd_len);
-		switch (handle_error(sock_info, len, lum->u.var_len_cmd.cmd_len, "counter event", &ret)) {
-		case MSG_OK:
-			break;
-		case MSG_ERROR:		/* Fallthrough */
-		case MSG_SHUTDOWN:
+		ret = copy_payload(payload, payload_size,
+				&var_len_cmd_data);
+
+		if (ret)
 			goto error;
-		}
-		args.counter_event.len = lum->u.var_len_cmd.cmd_len;
+
+		args.counter_event.len = payload_size;
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
 					(unsigned long) var_len_cmd_data,
@@ -1459,16 +1533,13 @@ int handle_message(struct sock_info *sock_info,
 #endif	/* CONFIG_LTTNG_UST_EXPERIMENTAL_COUNTER */
 	case LTTNG_UST_ABI_EVENT_NOTIFIER_CREATE:
 	{
-		len = ustcomm_recv_var_len_cmd_from_sessiond(sock,
-				&var_len_cmd_data, lum->u.var_len_cmd.cmd_len);
-		switch (handle_error(sock_info, len, lum->u.var_len_cmd.cmd_len, "event notifier", &ret)) {
-		case MSG_OK:
-			break;
-		case MSG_ERROR:		/* Fallthrough */
-		case MSG_SHUTDOWN:
+		ret = copy_payload(payload, payload_size,
+				&var_len_cmd_data);
+
+		if (ret)
 			goto error;
-		}
-		args.event_notifier.len = lum->u.var_len_cmd.cmd_len;
+
+		args.event_notifier.len = payload_size;
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
 					(unsigned long) var_len_cmd_data,
@@ -1481,8 +1552,8 @@ int handle_message(struct sock_info *sock_info,
 	default:
 		if (ops->cmd)
 			ret = ops->cmd(lum->handle, lum->cmd,
-					(unsigned long) &lum->u,
-					&args, sock_info);
+				(unsigned long) payload,
+				&args, sock_info);
 		else
 			ret = -ENOSYS;
 		break;
@@ -1492,15 +1563,27 @@ int handle_message(struct sock_info *sock_info,
 
 	if (ret >= 0) {
 		switch (lum->cmd) {
-		case LTTNG_UST_ABI_TRACER_VERSION:
-			lur.u.version = lum->u.version;
-			break;
-		case LTTNG_UST_ABI_TRACEPOINT_LIST_GET:
-			memcpy(&lur.u.tracepoint, &lum->u.tracepoint, sizeof(lur.u.tracepoint));
+		case LTTNG_UST_ABI_TRACER_VERSION: {
+                        /*
+			 * Payload was filled with the tracer version with the object callback.
+			 *
+			 * Even if the sent payload size is 0, the allocated
+			 * buffer for the payload is on the stack and can be
+			 * used to get the necessary information.
+			 */
+			lur.cmd.version = *(struct lttng_ust_abi_tracer_version *)payload;
 			break;
 		}
+		case LTTNG_UST_ABI_TRACEPOINT_LIST_GET: {
+			/*
+			 * Same rationale has with tracer version.
+			 */
+			lur.cmd.tracepoint = *(struct lttng_ust_abi_tracepoint_iter *)payload;
+			break;
+		}
+		}
 	}
-	DBG("Return value: %d", lur.ret_val);
+	DBG("Return value: %d", lur.header.ret_val);
 
 	ust_unlock();
 
@@ -1526,7 +1609,7 @@ int handle_message(struct sock_info *sock_info,
 	 * LTTNG_UST_TRACEPOINT_FIELD_LIST_GET needs to send the field
 	 * after the reply.
 	 */
-	if (lur.ret_code == LTTNG_UST_OK) {
+	if (lur.header.ret_code == LTTNG_UST_OK) {
 		switch (lum->cmd) {
 		case LTTNG_UST_ABI_TRACEPOINT_FIELD_LIST_GET:
 			len = ustcomm_send_unix_sock(sock,
@@ -1902,6 +1985,115 @@ error:
 	return;
 }
 
+static void
+flush_message(int sock, const struct ustcomm_ust_msg_header *lum, char *payload_buf)
+{
+	struct ustcomm_ust_reply lur = {};
+	struct msghdr msghdr;
+	struct iovec iov[] = {
+		{
+			.iov_base = payload_buf,
+			.iov_len = lum->payload_size,
+		},
+	};
+	ssize_t len;
+	int ret;
+
+	prepare_cmd_reply(&lur, lum->handle, lum->cmd, -ENOSYS);
+
+	if (!lum->payload_size)
+		goto no_payload;
+
+	memset(&msghdr, 0, sizeof(msghdr));
+
+	msghdr.msg_iov = iov;
+	msghdr.msg_iovlen = LTTNG_ARRAY_SIZE(iov);
+	msghdr.msg_control = NULL;
+	msghdr.msg_controllen = lum->ancillary_size;
+
+	len = recvmsg(sock, &msghdr, 0);
+
+	EXPECT_PAYLOAD_SIZE(lum->payload_size, len) {
+		ERR("error while flushing unknown command");
+	}
+
+no_payload:
+	ret = send_reply(sock, &lur);
+
+	if (ret != 0)
+		ERR("failed to send reply while flushing unknown command: %d", ret);
+}
+
+static int read_message_payload_and_ancillary(int sock,
+					uint32_t expected_payload_size,
+					uint32_t expected_ancillary_size,
+					char *payload_buf, uint32_t payload_buf_size,
+					char *ancillary_buf, uint32_t ancillary_buf_size)
+{
+	struct msghdr msghdr;
+	ssize_t size_recv;
+	struct cmsghdr *cmptr;
+	struct iovec iov[] = {
+		{
+			.iov_base = payload_buf,
+			.iov_len = expected_payload_size,
+		},
+	};
+
+	if (!expected_payload_size)
+		return 0;
+
+	if (expected_payload_size > payload_buf_size) {
+		ERR("message payload of `%" PRIu32 "` exceed maximum payload size `%" PRIu32 "`\n",
+			expected_payload_size, payload_buf_size);
+		return -EBADMSG;
+	}
+
+	if (expected_ancillary_size > ancillary_buf_size) {
+		ERR("message ancillary of `%" PRIu32 "` exceed maximum ancillary size `%" PRIu32 "`\n",
+			expected_ancillary_size, ancillary_buf_size);
+		return -EBADMSG;
+	}
+
+	memset(&msghdr, 0, sizeof(msghdr));
+
+	msghdr.msg_iov = iov;
+	msghdr.msg_iovlen = LTTNG_ARRAY_SIZE(iov);
+	msghdr.msg_control = ancillary_buf;
+	msghdr.msg_controllen = ancillary_buf_size;
+
+	size_recv = recvmsg(sock, &msghdr, MSG_CMSG_CLOEXEC);
+
+	if (size_recv != expected_payload_size) {
+		switch (size_recv) {
+		case -1:
+			return -errno;
+		case 0:
+			return -ESHUTDOWN;
+		default:
+			/*
+			 * No error nor shutdown, but the received payload size
+			 * is not matching what is described in the message header.
+			 */
+			return -EBADMSG;
+		}
+	}
+
+	if (expected_ancillary_size) {
+		cmptr = CMSG_FIRSTHDR(&msghdr);
+
+		if (!cmptr)
+			return -EBADMSG;
+		EXPECT_ANCILLARY_SIZE(cmptr->cmsg_len, CMSG_LEN(expected_ancillary_size)) {
+			return -EBADMSG;
+		}
+
+		memmove(ancillary_buf, CMSG_DATA(cmptr), expected_ancillary_size);
+	}
+
+	return 0;
+}
+
 /*
  * This thread does not allocate any resource, except within
  * handle_message, within mutex protection. This mutex protects against
@@ -2148,7 +2340,14 @@ restart:
 
 	for (;;) {
 		ssize_t len;
-		struct ustcomm_ust_msg lum;
+		struct ustcomm_ust_msg_header lum;
+		char payload_buf[LTTNG_UST_COMM_MAX_PAYLOAD_SIZE];
+		char ancillary_buf[CMSG_SPACE(LTTNG_UST_COMM_MAX_SEND_FDS * sizeof(int))]
+			__attribute__((aligned(__alignof__(struct cmsghdr))));
+		bool known;
+
+		memset(payload_buf, 0, sizeof(payload_buf));
+		memset(ancillary_buf, 0, sizeof(ancillary_buf));
 
 		len = ustcomm_recv_unix_sock(sock, &lum, sizeof(lum));
 		switch (len) {
@@ -2172,16 +2371,37 @@ restart:
 			ust_unlock();
 			goto end;
 		case sizeof(lum):
-			print_cmd(lum.cmd, lum.handle);
-			ret = handle_message(sock_info, sock, &lum);
+			known = print_cmd(lum.cmd, lum.handle,
+					lum.payload_size, lum.ancillary_size);
+			if (!known) {
+				flush_message(sock, &lum, payload_buf);
+				continue;
+			}
+
+			ret = read_message_payload_and_ancillary(sock,
+								lum.payload_size, lum.ancillary_size,
+								payload_buf, sizeof(payload_buf),
+								ancillary_buf, sizeof(ancillary_buf));
+			if (ret) {
+				ERR("Error %d while reading message payload for %s socket",
+					-ret,
+					sock_info->name);
+				goto end;
+			}
+			ret = handle_message(sock_info, sock, &lum,
+					payload_buf, lum.payload_size,
+					ancillary_buf, lum.ancillary_size);
 			if (ret) {
 				ERR("Error handling message for %s socket",
 					sock_info->name);
+
 				/*
 				 * Close socket if protocol error is
 				 * detected.
 				 */
-				goto end;
+				if (ret != -ENOSYS)
+					goto end;
+
 			}
 			continue;
 		default:
